from ..errors import PagesOutOfRange
from .wrap_tools import make_repr

def paginate(iterable, per_page: int = 10):
    """A function that paginates the given iterable.

    Simply wraps iterable in :class:`.Paginator` and returns it.

    Parameters
    ----------
    iterable: Sequence[`Any`]
        A finite sequence to paginate.

    per_page: :class:`int`
        Amount of elements in iterable to present on each page.

    Returns
    -------
    :class:`.Paginator`
        A paginator that has an ``iterable`` wrapped.
    """
    return Paginator(to_paginate=iterable, per_page=per_page)

class Paginator:
    """A class that implements a pagination system.

    Parameters
    ----------
    to_paginate: Sequence[Any]
        A sequence to paginate, such that list(to_paginate) can be performed.

    per_page: :class:`int`
        Number of elements each page can have.
    """

    def __init__(self, to_paginate, per_page: int = 10):
        self._list = list(to_paginate)
        self._per_page = per_page
        self._current_page = 0
        self.reload()

    def _configure_pages(self):
        y = self._length / self._per_page
        x = round(y)

        self._pages = x+1 if x < y else x

    def __str__(self):
        return self.current_state

    def __repr__(self):
        info = {
            'can_run': self.can_run(),
            'length': self.length,
            'per_page': self.per_page,
            'pages': self.get_pages_count(),
            'current': self.get_curr_page()
        }
        return make_repr(self, info)

    def __getitem__(self, index: int):
        return self._list[index]

    def __setitem__(self, index: int, item):
        self._list[index] = item
        self.reload()

    @property
    def length(self):
        """:class:`int`: Total amount of elements in a paginator."""
        return self._length

    @property
    def per_page(self):
        """:class:`int`: Amount of elements shown on each page."""
        return self._per_page

    @property
    def list(self):
        """List[`Any`] Paginated sequence, represented as list."""
        return self._list

    @property
    def current_state(self):
        """:class:`str`: Returns a string representation of the Paginator.

        Theoretically, this is only used when printing the Paginator's state.
        """
        p = self.get_curr_page()
        per_page = self.per_page

        _from, _to = (p*per_page, (p+1)*per_page)

        return (
            '[gd.Paginator]\n'
            '[Can_Run: {can_run}]\n'
            '[Length: {length}]\n[Pages: {pages_count}]\n'
            '[Current_State]\n[Page: {page}]\n'
            '[Showing: {_from} to {_to} of {length}]'
        ).format(
            can_run=self.can_run(), length=self.length,
            pages_count=self.get_pages_count(),
            page=self.get_curr_page(), _from=_from, _to=_to
        )

    def load_from(self, iterable):
        """A function that paginates the given iterable.

        Basically reloads already created paginator with given iterable.

        Parameters
        ----------
        iterable: Sequence[`Any`]
            A finite sequence to paginate.

        per_page: :class:`int`
            Amount of elements in iterable to present on each page.
        """
        self._list = list(iterable)
        self.reload()

    def reload(self):
        """Reconfigure length and configure pages amount."""
        self._length = len(self._list)
        self._configure_pages()

    def append(self, obj):
        """Append ``obj`` to a paginated sequence, then reload."""
        self._list.append(obj)
        self.reload()

    def pop(self, index: int = -1):
        """Pops out an element.

        Deletes element with given ``index``, and returns it.
        Reloads paginator as well.
        """
        value = self._list.pop(index)
        self.reload()
        return value

    def clear(self):
        """Clears :attr:`Paginator.list`, then reloads."""
        self._list.clear()
        self.reload()

    def get_pages_count(self):
        """:class:`int`: Get current amount of pages."""
        return self._pages

    def get_curr_page(self):
        """:class:`int`: Get a number of the current page."""
        return self._current_page

    def has_next(self):
        """:class:`bool`: Indicates whether does the next page exists."""
        return True if self.get_curr_page() < self.get_pages_count() else False

    def has_prev(self):
        """:class:`bool`: Indicates if previous page is existing."""
        return True if self.get_curr_page() > 0 else False

    def can_run(self):
        """:class:`bool`: Indicates whether paginator is empty."""
        return self.length > 0

    def move_to_next(self):
        """Tries to go to the next page.

        Raises
        ------
        :exc:`.PagesOutOfRange`
            The next page does not exist.
        """
        if not self.has_next():
            raise PagesOutOfRange(
                page = self.get_curr_page() + 1,
                info = self.get_pages_count()
            )
        self._current_page += 1

    def move_to_prev(self):
        """Attempts to go to the previous page.

        Raises
        ------
        :exc:`.PagesOutOfRange`
            The previous page does not exist.
        """
        if not self.has_prev():
            raise PagesOutOfRange(
                page = self.get_curr_page() - 1,
                info = 'Page does not support negative integers.'
            )
        self._current_page -= 1

    def move_to(self, page: int):
        """Moves to the ``page`` given.

        Parameters
        ----------
        page: :class:`int`
            A page to move to.

        Raises
        ------
        :exc:`.PagesOutOfRange`
            Given ``page`` does not exist.
        """
        page = int(page)
        if abs(page) > page:
            raise PagesOutOfRange(
                page = page,
                info = 'Page does not support negative integers.'
            )
        if page > self.get_pages_count():
            raise PagesOutOfRange(
                page = page,
                info = self.get_pages_count()
            )
        self._current_page = page

    def view_page(self, page: int = None):
        """View elements on a page.

        Parameters
        ----------
        page: :class:`int`
            If ``page`` is ``None``, returns elements on a current page.
            Otherwise, tries to go to the ``page``, and returns its elements.
            (without switching to that page)

        Raises
        ------
        :exc:`.PagesOutOfRange`
            Raised when given page is out of range.
        """
        if page is None:
            page_to_view = self.get_curr_page()
            to_pass = page_to_view * self.per_page
            res = self.list[to_pass:(to_pass + self.per_page)]
            return res

        else:
            page = int(page)
            if abs(page) > page:
                raise PagesOutOfRange(
                    page = page,
                    info = "Page does not support negative integers."
                )
            if page > self.get_pages_count():
                raise PagesOutOfRange(
                    page = page,
                    info = self.get_pages_count()
                )
            else:
                to_pass = page * self.per_page
                res = self.list[to_pass:(to_pass + self.per_page)]
                return res

    def get_all(self):
        """Pretty similar to self.list, but returns List[List[`Any`]], where each
        inner list represents one page.
        """
        return [self.view_page(i) for i in range(self.get_pages_count())]



'''A module for generating Geometry Dash Icon Set images.
Credits to Alex1304 for making the SpriteFactory in Java,
which is mostly translated by NeKitDS to Python.
[NOT WORKING YET]
'''
import ast
import pkg_resources
import re

from typing import List, Callable

from xml.etree import ElementTree

from PIL import Image, ImageOps

from .sprite import Sprite, Rectangle
from ..colors import Color, colors
from ..iconset import IconSet

main_path = 'resources/GJ_GameSheet02-uhd'

test_path = pkg_resources.resource_filename(__name__, 'resources/result.png')

def make_path(file_ext):
    return pkg_resources.resource_filename(__name__, main_path + file_ext)


class SpriteGenUtil:
    def __init__(self):
        self.plist_path = make_path('.plist')
        self.icon_types = ('cube', 'ship', 'ball', 'ufo', 'wave', 'robot', 'spider')

    def parse_xml(self):  # parsing xml into nice python dict
        parsed = ElementTree.parse(self.plist_path)
        load_dict = {}

        something = parsed.getroot()

        for i in range(2):  # going deeper into the tree
            something = something.getchildren()[i]

        for j in range(0, len(something.getchildren()), 2):

            temp = something[j].text
            start = something[j+1].getchildren()

            xy, wh = parse_array(start[9])

            formatted = {
                'spriteOffset': parse_array(start[3]),
                'spriteSize': parse_array(start[5]),
                'spriteSourceSize': parse_array(start[7]),
                'spriteRect': Rectangle(*xy, *wh),
                'isRotated': parse_bool(start[11])
            }
            load_dict[temp] = formatted

        return load_dict

    def find_bases(self, iconset: IconSet):  # finding bases for further creating
        final = []
        for icon in self.icon_types:
            temp = fix_int(getattr(iconset, icon))
            prefix = fix_prefix(icon)
            base = '{}_{}_..._001.png'.format(prefix, temp)
            final.append(base)
        return final

    def format_related(self, parsed_xml, base_list):  # getting all sprites related to bases
        final = {}
        for base in base_list:
            start = base.split('...')[0]
            temp = [key for key in parsed_xml.keys() if (key.startswith(start))]
            icon_type = self.icon_types[base_list.index(base)]
            final[icon_type] = temp
        return final

    def to_sprite_object(self, name, info_dict):  # converting dict to gd.graphics.Sprite
        sprite = None
        id_idx = (1 if 'ball' not in name else 2)
        if info_dict:
            sprite = Sprite(
                name = name.split('.')[0], id = int(name.split('_')[id_idx]),
                offset = info_dict.get('spriteOffset'),
                size = info_dict.get('spriteSize'),
                source_size = info_dict.get('spriteSourceSize'),
                upleft_corner = info_dict.get('spriteUpLeftCorner'),
                rectangle = info_dict.get('spriteRect'),
                is_rotated = info_dict.get('isRotated')
            )
        return sprite

    def retrieve_sprites(self, parsed_xml, icon_type, related_dict):  # getting a list of gd.graphics.Sprite objects
        related_list = related_dict.get(icon_type)
        store = []
        for element in related_list:
            temp = parsed_xml.get(element)
            sprite = self.to_sprite_object(element, temp)
            store.append(sprite)
        return store

class SpriteGen:
    def __init__(self):
        self.image_path = make_path('.png')

    def rotate(self, image, deg):
        deg = deg % 360 + (-sign(deg) * 360 if abs(deg) > 180 else 0)
        new = image.rotate(deg, expand=True)
        return new

    def reduce_brightness(self, image, sprite):
        pixels = image.load()
        x, y = image.size

        for i in range(x):
            for j in range(y):
                rgba = pixels[i, j]

                color_rgba = Color(0x808080).to_rgba()
                gen = (a&b for a, b in zip(rgba, color_rgba))

                pixels[i, j] = tuple(gen)

    def fix_brightness(self, image, sprite):
        if '001D' in sprite.name and '_glow_' not in sprite.name:
            self.reduce_brightness(image, sprite)

    def load_image(self):  # loading sprite sheet image
        image = Image.open(self.image_path)
        return image

    def apply_color(self, pixel, color):
        res = []
        for a, b in zip(pixel, color.to_rgba()):
            res.append(
                int(b/(255/a)) if a > 20 else a
            )
        return tuple(res)

    def make_blank(self):
        t = (255, 255, 255, 0)
        image = Image.new('RGBA', (250, 250), t)
        return image

    def get_sprite_image(self, sheet, sprite):
        rect = sprite.get_rectangle()
        x, y = rect.get_coords()
        w, h = rect.size

        sx, sy = sheet.size

        if 'robot' in sprite.name:
            fix_robot_offset(sprite)

        rw, rh = (
            (w, h) if not sprite.is_rotated() else (h, w)
        )

        border = (x, y, sx-(x+rw), sy-(y+rh))

        print(border)

        image = ImageOps.crop(sheet, border)

        print(image)

        if hasattr(sprite, 'fix_rotation_deg'):
            image = self.rotate(image, sprite.fix_rotation_deg)

        if sprite.is_rotated():
            image = self.rotate(image, 90)

        self.fix_brightness(image, sprite)

        return image

    def color(self, img, color):
        pixels = img.load()
        x, y = img.size
        for i in range(x):
            for j in range(y):
                pixel = pixels[i, j]
                pixels[i, j] = self.apply_color(pixel, color)
        return img

def gen_center_offset(main, to_paste, sprite):
    name = sprite.name

    center_x, center_y = (n//2 for n in sprite.size)
    off_x, off_y = sprite.offset

    draw_x = 100 - center_x + off_x
    draw_y = 100 - center_y - off_y

    cases = {
        fix_prefix('robot') in name: (0, -20),
        fix_prefix('spider') in name: (6, -5),
        fix_prefix('ufo') in name: (0, 30)
    }

    draw_off_x, draw_off_y = cases.get(True, (0, 0))

    x, y = (25 + draw_off_x + draw_x), (25 + draw_off_y + draw_y)
    return x, y

def fix_glow_color(colors):
    color1, color2 = colors
    if color2.index == 15:
        color2 = colors[12] if color1.index == 15 else color1
    return color2

def predict_color(sprite, colors):
    color1, color2 = colors
    if '_glow_' in sprite.name:
        return fix_glow_color(colors)
    elif '_2_001' in sprite.name:
        return color2
    elif all(part not in sprite.name for part in ('extra', '_3_001')):
        return color1

def fix_robot_offset(sprite) -> None:
    id = sprite.id
    name = sprite.name
    offX, offY = (0, 0)
    if str(id) + '_02_' in name:
        sprite.fix_rotation_deg = -45
        offX -= (50 if '001D' in name else 40)
        offY -= 20
        if '_2_' in name:
            cases = {
                id in (2, 5, 6, 8, 9, 11, 12, 15, 17, 24): (15, -5),
                id in (7, 10, 19, 20): (0, 7),
                id == 13: (10, -4), id == 18: (-1, -1),
                id in (21, 25): (12, 0), id == 22: (0, -5),
                id in (3, 26): (1, 0), id == 23: (-3, -2)
            }
            offTX, offTY = cases.get(True, (0, 0))
            offX += offTX
            offY += offTY
    elif str(id) + '_03_' in name:
        sprite.fix_rotation_deg = 45
        offX -= (40 if '001D' in name else 30)
        offY -= (52 if (id == 21 and '_2_' not in name) else 60)
    elif str(id) + '_04_' in name:
        offX -= (10 if '001D' in name else 0)
        offY -= 70
    print(offX, offY)
    sprite.update_offset(offX, offY)

def fix_spider_offset(sprite):
    pass

def reorder(splist):
    splist.reverse()
    push_sprite_to_back_if(splist, lambda sp: '_2_' in sp.name)
    pull_sprite_to_front_if(
        splist, lambda sp: re.match('(robot|spider)_[0-9]{2,3}_(02|03|04)_.*', sp.name) is not None
    )
    dupe_sprite_if(
        splist, lambda sp: re.match('robot_[0-9]{2,3}_(02|03|04)_.*', sp.name) is not None, 1
    )
    dupe_sprite_if(
        splist, lambda sp: re.match(
            'spider_[0-9]{2,3}_02_.*', sp.name) is not None and not 'extra' in sp.name, 2
    )
    pull_sprite_to_front_if(
        splist, lambda sp: re.match(
            'robot_[0-9]{2,3}_02_.*', sp.name) is not None and not sp.name.endswith('D')
    )
    pull_sprite_to_front_if(
        splist, lambda sp: re.match(
            'robot_[0-9]{2,3}_04_.*', sp.name) is not None and not sp.name.endswith('D')
    )
    push_sprite_to_back_if(splist, lambda sp: '_2_' not in sp.name and sp.name.endswith('D'))
    push_sprite_to_back_if(splist, lambda sp: '_2_' in sp.name and sp.name.endswith('D'))
    push_sprite_to_back_if(
        splist, lambda sp: re.match('spider_[0-9]{2,3}_04_.*', sp.name) is not None
    )
    pull_sprite_to_front_if(splist, lambda sp: 'extra' in sp.name)
    push_sprite_to_back_if(splist, lambda sp: '_glow_' in sp.name)
    push_sprite_to_back_if(splist, lambda sp: '_3_' in sp.name and 'bird' in sp.name)

def pull_sprite_to_front_if(splist: List[Sprite], predicate: Callable[[Sprite], bool]):
    offset = 0
    for i in range(len(splist)):
        sprite = splist[i-offset]
        if predicate(sprite):
            splist.pop(i-offset)
            splist.append(sprite)
            offset += 1

def push_sprite_to_back_if(splist: List[Sprite], predicate: Callable[[Sprite], bool]):
    for i in range(len(splist)):
        sprite = splist[i]
        if predicate(sprite):
            splist.pop(i)
            splist.insert(0, sprite)

def dupe_sprite_if(
    splist: List[Sprite], predicate: Callable[[Sprite], bool], dupe_n: int, front: bool = False
):
    ssize = len(splist)
    offset = 0
    for i in range(ssize):
        sprite = splist[i+offset]
        if predicate(sprite):
            dupe = sprite.duplicate()
            for i in range(dupe_n):
                if front:
                    splist.append(dupe)
                else:
                    splist.insert(0, dupe)
                    offset += 1
                dupe = dupe.duplicate()

def sign(number):
    return (-1 if number < 0 else (1 if number > 0 else 0))

def parse_array(element, deep=-1):
    ret = ast.literal_eval(
        element.text.replace('{', '(').replace('}', ')')
    )
    if deep > -1:
        ret = ret[deep]
    return ret

def parse_bool(element):
    return bool(element.tag.lower().replace('false', ''))  # this will eventually give us real bools from strings

def fix_prefix(prefix):
    cases = {
        'cube': 'player',
        'ball': 'player_ball',
        'ufo': 'bird',
        'wave': 'dart'
    }
    return cases.get(prefix, prefix)

def fix_int(n):
    return '{:02}'.format(n) if n else n



import logging
from io import BytesIO

from PIL import Image, ImageOps

from ._async import run_blocking_io
from .wrap_tools import benchmark
from ..errors import FailedCaptcha

border = (5, 8, 16, 7)  # borders to crop
# we recieve picture 65x25 from gd server (1625 pixels to iterate through)
# and then we can cut borders, getting 44x10 image (440 pixels to iterate through)
# so, amount of iterations is ~3.693 times lower.

log = logging.getLogger(__name__)

class Captcha:
    @classmethod
    @benchmark
    def solve(cls, image_bytes: bytes, should_log: bool = False):
        """Solves GD Captcha from given buffer/bytes.

        Method: get_image -> crop -> walk through pixels
        -> mark bright pixels -> predict digits from existing setup
        -> connect digits -> return

        This is used to make operations like changing username possible.

        Parameters
        ----------
        image_bytes: :class:`bytes`
            Bytes to get the Captcha image from.
            Converted to BytesIO to make initialization of image
            a little bit faster.

        should_log: :class:`bool`
            Indicates if message about solved Captcha should be logged.

        Returns
        -------
        :class:`int`
            Code that is shown in Captcha.
        """
        # get image from BytesIO, then crop the borders
        img = ImageOps.crop(Image.open(BytesIO(image_bytes)), border)
        pixels = img.load()
        predict = cls.walk_through(pixels, img.size)
        code = cls.connect_result(predict)
        if should_log:
            log.debug("Solved a Captcha. The code was %s.", code)
        return code

    @classmethod
    async def aio_solve(cls, image_bytes: bytes, should_log: bool = False):
        """|coro|

        Solves a Captcha asynchronously in a new event loop.

        Basically runs :meth:`.Captcha.solve` with :meth:`asyncio.AbstractEventLoop.run_in_executor`.

        Passes ``None`` to that method as a first argument.
        """
        return await run_blocking_io(cls.solve, image_bytes, should_log)

    @classmethod
    def walk_through(cls, pixel_map, size):
        """The main body of Captcha solving.

        Walks through pixels, finds bright ones, stores them,
        tries to predict each digit.

        Parameters
        ----------
        pixel_map: :class:`PIL.PixelAccess`
            A map of pixels to walk through.

        size: Tuple[:class:`int`, :class:`int`]
            A tuple representing width and height of the image.

        Returns
        -------
        :class:`list`
            A list, containing five digits to solve the Captcha.
        """
        lst = [[] for _ in range(5)]  # store
        g = 0
        final_res = []
        x, y = size
        for i in range(x):  # for every column
            temp = []
            for j in range(y):  # for every row
                rgb = pixel_map[i, j]
                if all(rgb[i] > 200 for i in range(3)):  # if pixel is white
                    temp.append(j)  # append 'index' in column of pixel
            if not temp or (i == x-1):  # if temp is empty or index is last
                if lst[g]:  # if we actually have something to predict
                    predicted = cls.predict_digit(lst[g])  # predict from list
                    final_res.append(predicted)
                    g += 1  # move to next list element
            else:  # otherwise
                s = lst[g]
                if len(s) < 2:  # if less than two rows
                    s.append(temp)  # append
        return final_res

    @classmethod
    def predict_digit(cls, res):
        cases = cls.init_numbers()
        try:
            return cases.index(res)
        except ValueError:
            raise FailedCaptcha('Unknown result for digit was recieved: {}.'.format(res)) from None

    @classmethod
    def init_numbers(cls):
        """Get bright pixel index setups for number predicting."""
        setup = [
            [[3, 4, 5, 6], [2, 3, 4, 5, 6, 7]],
            [[2, 9], [1, 2, 9]],
            [[2, 9], [1, 2, 8, 9]],
            [[1, 8], [0, 1, 8, 9]],
            [[5, 6], [4, 5, 6]],
            [[0, 1, 2, 3, 4, 7], [0, 1, 2, 3, 4, 7, 8]],
            [[2, 3, 4, 5, 6, 7], [1, 2, 3, 4, 5, 6, 7, 8]],
            [[0, 8, 9], [0, 7, 8, 9]],
            [[2, 6, 7], [1, 2, 3, 5, 6, 7, 8]],
            [[2, 3], [1, 2, 3, 4, 7, 8]]
        ]
        return setup

    @classmethod
    def connect_result(cls, res):
        return int(str().join(map(str, res)))



import base64
import hashlib
import random
import string
import zlib

from typing import Sequence, Union

from ..mapper import mapper_util
from .xor_cipher import XORCipher as XOR


class Coder:
    keys = {
        'message': '14251',
        'levelpass': '26364',
        'accountpass': '37526',
        'levelscore': '39673',
        'level': '41274',
        'comment': '29481',
        'challenges': '19847',
        'rewards': '59182',
        'like_rate': '58281',
        'userscore': '85271'
    }
    salts = {
        'level': 'xI25fpAapCQg',
        'comment': 'xPT6iUrtws0J',
        'like_rate': 'ysg6pUrtjn0J',
        'userscore': 'xI35fsAapCRg',
        'levelscore': 'yPg6pUrtWn0J'
    }

    @classmethod
    def normal_xor(cls, data: bytes, key: int):
        """Applies simple XOR on an array of bytes.

        Parameters
        ----------
        data: :class:`bytes`
            Data to apply XOR on.

        key: :class:`int`
            A key to XOR data with.

        Returns
        -------
        :class:`str`
            Decoded data as a string.
        """
        return bytearray(i^key for i in data).decode()

    @classmethod
    def decode_gamesave(cls, save: str, needs_xor: bool = True):
        if needs_xor:
            save = cls.normal_xor(save.encode(), 11)
        data = mapper_util.normalize(save).encode()
        from_b64 = base64.b64decode(data)[10:]
        return zlib.decompress(from_b64, -zlib.MAX_WBITS).decode(errors='replace')

    @classmethod
    def gen_rs(cls, length: int = 10):
        """Generates a random string of required length.

        Uses [A-Z][a-z][0-9] character set.

        Parameters
        ----------
        length: :class:`int`
            Amount of characters for a string to have.

        Returns
        -------
        :class:`str`
            Generated string.
        """
        sset = string.ascii_letters + string.digits
        rand = random.choices(sset, k=length)
        final = ''.join(rand)
        return final

    @classmethod
    def encode(cls, type: str, string: str):
        """Encodes a string, combining XOR and Base64 encode methods.

        Used in different aspects of gd.py.

        Parameters
        ----------
        type: :class:`str`
            String representation of type, e.g. ``'levelpass'``.
            Used to define a XOR key.

        string: :class:`str`
            String to encode.

        Returns
        -------
        :class:`str`
            Encoded string.
        """
        ciphered = XOR.cipher(key=cls.keys[type], string=string)
        encoded = base64.b64encode(ciphered.encode()).decode()
        return encoded

    @classmethod
    def decode(cls, type: str, string: str):
        """Decodes a XOR -> Base64 ciphered string.

        .. note::
            Due to the fact that decode and encode work almost the same,
            the following is true:

            .. code-block:: python3

                Coder.decode('message', Coder.encode('message', 'NeKit')) == 'NeKit'  # True

        Parameters
        ----------
        type: :class:`str`
            String representation of a type, e.g. ``'level'``.
            Used to define a XOR key.

        string: :class:`str`
            String to decode.

        Returns
        -------
        :class:`str`
            Decoded string.
        """
        ciphered = base64.b64decode(string).decode()
        decoded = XOR.cipher(key=cls.keys[type], string=ciphered)
        return decoded

    @classmethod
    def gen_chk(cls, type: str, values: Sequence[Union[int, str]]):
        """Generates a "chk" value, used in different requests to GD servers.

        The method is: combine_values -> add salt -> sha1 hash
        -> XOR -> Base64 encode -> return

        Parameters
        ----------
        type: :class:`str`
            String representation of type, e.g. ``'comment'``.
            Used to define salt and XOR key.

        values: Sequence[Union[:class:`int`, :class:`str`]]
            Sequence of values to generate a chk with.

        Returns
        -------
        :class:`str`
            Generated ``'chk'``, represented as string.
        """
        values = list(map(str, values))
        salt = cls.salts.get(type, '')  # get salt
        string = ''
        # combine
        # special case for 'levelscore'
        if (type == 'levelscore'):
            for i in range(len(values)-1):
                string += values[i]
                if (i == 7):
                    string += '1'
            string += salt + values[-1]
        # normal case
        else:
            for value in values:
                string += value
            string += salt
        # sha1 hash
        hashed = hashlib.sha1(string.encode()).hexdigest()
        # XOR
        xored = XOR.cipher(key=cls.keys[type], string=hashed)
        # Base64
        final = base64.b64encode(xored.encode()).decode()
        return final

    @classmethod
    def unzip(cls, string: str):
        """zlib decompresses a level string.

        Used to unzip level data.

        Parameters
        ----------
        string: :class:`str`
            String to unzip, encoded in Base64.

        Returns
        -------
        :class:`str`
            Unzipped level data, as string.
        """
        decoded = base64.b64decode(mapper_util.normalize(string).encode())

        try:
            unzipped = zlib.decompress(decoded, zlib.MAX_WBITS)
        except zlib.error:
            unzipped = zlib.decompress(decoded[10:], -zlib.MAX_WBITS)

        final = unzipped.decode(errors='replace')
        if final.endswith(';'):
            final = final[:-1]

        return final

    @classmethod
    def gen_level_lb_seed(cls, jumps: int = 0, percentage: int = 0, seconds: int = 0):
        return 1482 + (jumps + 3991) * (percentage + 8354) + ((seconds + 4085)**2) - 50_028_039



from .abstractentity import AbstractEntity
from .utils.http_request import http
from .utils.wrap_tools import make_repr

class Song(AbstractEntity):
    """Class that represents Geometry Dash/Newgrounds songs.
    This class is derived from :class:`.AbstractEntity`.
    """
    def __init__(self, **options):
        super().__init__(**options)
        self.options = options

    def __repr__(self):
        info = {
            'id': self.id,
            'name': repr(self.name),
            'author': self.author
        }
        return make_repr(self, info)

    @property
    def id(self):
        """:class:`int`: An ID of the song."""
        return self.options.get('id')

    @property
    def name(self):
        """:class:`str`: A name of the song."""
        return self.options.get('name')

    @property
    def size(self):
        """:class:`float`: A float representing size of the song, in megabytes."""
        return self.options.get('size')

    @property
    def author(self):
        """:class:`str`: An author of the song."""
        return self.options.get('author')

    @property
    def link(self):
        """:class:`str`: A link to the song on Newgrounds, e.g. ``.../audio/listen/id``."""
        return self.options.get('links')[0]

    @property
    def dl_link(self):
        """:class:`str`: A link to download the song, used in :meth:`.Song.download`."""
        return self.options.get('links')[1]

    def is_custom(self):
        """:class:`bool`: Indicates whether the song is custom or not."""
        return self.options.get('custom')

    async def download(self):
        """|coro|

        Download a song from Newgrounds.

        Returns
        -------
        :class:`bytes`
            A song as bytes.
        """
        resp = await http.normal_request(self.dl_link)
        return await resp.content.read()



import base64 as b64
import urllib.parse

from .song import Song
from .user import UserStats, User
from .colors import Colour, colors
from .comment import Comment
from .abstractuser import AbstractUser
from .level import Level
from .message import Message
from .iconset import IconSet
from .friend_request import FriendRequest
from .level_packs import MapPack, Gauntlet

from .utils import convert_to_type
from .utils.converter import Converter
from .utils.indexer import Index as i
from .utils.crypto.coders import Coder
from .utils.mapper import mapper_util
from .utils.routes import Route
from .utils.enums import (
    MessagePolicyType, FriendRequestPolicyType,
    CommentPolicyType, StatusLevel,
    LevelLength, TimelyType, CommentType,
    MessageOrRequestType, IconType, GauntletEnum
)

class ClassConverter:

    @classmethod
    def song_convert(cls, s):
        quoted_url = s[i.SONG_URL]
        dl_link = urllib.parse.unquote(quoted_url)
        return Song(
            name = str(s[i.SONG_TITLE]),
            author = str(s[i.SONG_AUTHOR]),
            id = s[i.SONG_ID],
            size = float(s[i.SONG_SIZE]),
            links = (
                Route.NEWGROUNDS_SONG_LISTEN + str(s[i.SONG_ID]),
                dl_link
            ),
            custom = True,
        )

    @classmethod
    def song_from_kwargs(cls, **kwargs):
        return Song(**kwargs)

    @classmethod
    def user_stats_convert(cls, s, client=None):
        return UserStats(
            account_id = s[i.USER_ACCOUNT_ID], name = str(s[i.USER_NAME]), id = s[i.USER_PLAYER_ID],
            stars = s[i.USER_STARS], demons = s[i.USER_DEMONS], cp = s[i.USER_CREATOR_POINTS],
            diamonds = s[i.USER_DIAMONDS], coins = s[i.USER_COINS], client = client,
            secret_coins = s[i.USER_SECRET_COINS], lb_place = s.get(i.USER_TOP_PLACE, -1)
        )

    @classmethod
    def user_convert(cls, s, client=None):
        dm = MessagePolicyType.from_value(s[i.USER_PRIVATE_MESSAGE_POLICY])
        fr_rq = FriendRequestPolicyType.from_value(s[i.USER_FRIEND_REQUEST_POLICY])
        comment = CommentPolicyType.from_value(s[i.USER_COMMENT_HISTORY_POLICY])

        stat = StatusLevel.from_value(s[i.USER_ROLE])
        rnk = s[i.USER_GLOBAL_RANK]
        rank = rnk if rnk else None

        youtube = str(s[i.USER_YOUTUBE])
        yt = (youtube, 'https://www.youtube.com/channel/' + youtube) if youtube else (None, None)

        twitter = str(s[i.USER_TWITTER])
        twt = (twitter, 'https://twitter.com/' + twitter) if twitter else (None, None)

        twitch = str(s[i.USER_TWITCH])
        twch = (twitch, 'https://twitch.tv/' + twitch) if twitch else (None, None)

        return User(
            name = str(s[i.USER_NAME]), id = s[i.USER_PLAYER_ID],
            stars = s[i.USER_STARS], demons = s[i.USER_DEMONS],
            secret_coins = s[i.USER_SECRET_COINS], coins = s[i.USER_COINS],
            cp = s[i.USER_CREATOR_POINTS], diamonds = s[i.USER_DIAMONDS],
            role = stat, global_rank = rank,
            account_id = s[i.USER_ACCOUNT_ID], youtube = yt,
            twitter = twt, twitch = twch,
            messages = dm, friend_requests = fr_rq, comments = comment,
            icon_setup = IconSet(
                main_icon = s[i.USER_ICON],
                color_1 = colors[s[i.USER_COLOR_1]],
                color_2 = colors[s[i.USER_COLOR_2]],
                main_icon_type = IconType.from_value(s[i.USER_ICON_TYPE]),
                has_glow_outline = bool(s[i.USER_GLOW_OUTLINE_2]),
                icon_cube = s[i.USER_ICON_CUBE],
                icon_ship = s[i.USER_ICON_SHIP],
                icon_ball = s[i.USER_ICON_BALL],
                icon_ufo = s[i.USER_ICON_UFO],
                icon_wave = s[i.USER_ICON_WAVE],
                icon_robot = s[i.USER_ICON_ROBOT],
                icon_spider = s[i.USER_ICON_SPIDER]
            ),
            client = client
        )


    @classmethod
    def level_convert(cls, s, song, creator, client=None):
        try:  # decode password
            password = Coder.decode(type='levelpass', string=s.get(i.LEVEL_PASS))
        except TypeError:
            password = None
        if password == '1':
            password = ''

        desc = b64.b64decode(mapper_util.normalize(s[i.LEVEL_DESCRIPTION])).decode()
        length = LevelLength.from_value(s[i.LEVEL_LENGTH])
        type = TimelyType.from_value(s[i.LEVEL_TIMELY_TYPE])
        game_version = s[i.LEVEL_GAME_VERSION]  # needs to be converted to represent real version.

        try:
            leveldata = Coder.unzip(s[i.LEVEL_DATA])
        except KeyError:  # level data not present
            leveldata = None

        diff = s[i.LEVEL_DIFFICULTY]
        demon_diff = s[i.LEVEL_DEMON_DIFFICULTY]
        is_demon = bool(s[i.LEVEL_IS_DEMON])
        is_auto = bool(s[i.LEVEL_IS_AUTO])
        difficulty = Converter.convert_level_difficulty(
            diff=diff, demon_diff=demon_diff, is_demon=is_demon, is_auto=is_auto)

        return Level(
            id = s[i.LEVEL_ID],
            name = str(s[i.LEVEL_NAME]),
            description = desc,
            version = s[i.LEVEL_VERSION],
            creator = creator,
            song = song,
            data = leveldata,
            is_demon = is_demon,
            is_auto = is_auto,
            difficulty = difficulty,
            stars = s[i.LEVEL_STARS],
            coins = s[i.LEVEL_COIN_COUNT],
            verified_coins = bool(s[i.LEVEL_COIN_VERIFIED]),
            is_epic = bool(s[i.LEVEL_IS_EPIC]),
            original = s[i.LEVEL_ORIGINAL],
            downloads = s[i.LEVEL_DOWNLOADS],
            rating = s[i.LEVEL_LIKES],
            score = s[i.LEVEL_FEATURED_SCORE],
            uploaded_timestamp = s.get(i.LEVEL_UPLOADED_TIMESTAMP),
            last_updated_timestamp = s.get(i.LEVEL_LAST_UPDATED_TIMESTAMP),
            length = length,
            game_version = game_version,
            stars_requested = s[i.LEVEL_REQUESTED_STARS],
            object_count = s[i.LEVEL_OBJECT_COUNT],
            type = type,
            time_n = s[i.LEVEL_TIMELY_INDEX],
            cooldown = s[i.LEVEL_TIMELY_COOLDOWN],
            client = client
        )

    @classmethod
    def map_pack_convert(cls, s, client=None):
        level_id_string = s.get(i.MAP_PACK_LEVEL_IDS, '0,0,0')
        level_ids = tuple(map(int, level_id_string.split(',')))

        color_string = s.get(i.MAP_PACK_COLOR, '255,255,255')
        color = Colour.from_rgb(*map(int, color_string.split(',')))

        difficulty = Converter.value_to_pack_difficulty(s.get(i.MAP_PACK_DIFFICULTY))

        return MapPack(
            id = s[i.MAP_PACK_ID],
            name = str(s[i.MAP_PACK_NAME]),
            level_ids = level_ids,
            stars = s[i.MAP_PACK_STARS],
            coins = s[i.MAP_PACK_COINS],
            difficulty = difficulty,
            color = color,
            client = client
        )

    @classmethod
    def gauntlet_convert(cls, s, client=None):
        level_id_string = s.get(i.GAUNTLET_LEVEL_IDS, '0,0,0,0,0')
        level_ids = tuple(map(int, level_id_string.split(',')))

        typeid = s[i.GAUNTLET_ID]

        name = GauntletEnum.from_value(typeid).desc + ' Gauntlet'

        return Gauntlet(
            id = typeid, name = name, level_ids = level_ids, client = client
        )

    @classmethod
    def message_convert(cls, s, s_2, client=None):
        _type = MessageOrRequestType.from_value(s[i.MESSAGE_INDICATOR])

        useful_dict = {
            'name': s[i.MESSAGE_SENDER_NAME],
            'id': s[i.MESSAGE_SENDER_ID],
            'account_id': s[i.MESSAGE_SENDER_ACCOUNT_ID]
        }
        user_1, user_2 = [
            ClassConverter.abstractuser_convert(elem, client) for elem in (useful_dict, s_2)
        ]

        is_normal = _type.value ^ 1

        subject = b64.b64decode(mapper_util.normalize(s[i.MESSAGE_SUBJECT])).decode()
        return Message(
            id = s[i.MESSAGE_ID],
            timestamp = s[i.MESSAGE_TIMESTAMP],
            subject = subject,
            is_read = bool(s[i.MESSAGE_IS_READ]),
            author = user_1 if is_normal else user_2,
            recipient = user_2 if is_normal else user_1,
            type = _type,
            client = client
        )

    @classmethod
    def request_convert(cls, s, s_2, client=None):
        _type = MessageOrRequestType.from_value(s[i.REQUEST_INDICATOR])
        useful_dict = {
            'name': s[i.REQUEST_SENDER_NAME],
            'id': s[i.REQUEST_SENDER_ID],
            'account_id': s[i.REQUEST_SENDER_ACCOUNT_ID]
        }

        is_normal = _type.value ^ 1
        user_1, user_2 = (ClassConverter.abstractuser_convert(elem, client) for elem in (useful_dict, s_2))

        return FriendRequest(
            id = s[i.REQUEST_ID],
            timestamp = s[i.REQUEST_TIMESTAMP],
            body = b64.b64decode(mapper_util.normalize(s[i.REQUEST_BODY])).decode(),
            is_read = True if bool(s[i.REQUEST_STATUS]) ^ 1 else False,
            author = user_1 if is_normal else user_2,
            recipient = user_2 if is_normal else user_1,
            type = _type,
            client = client
        )

    @classmethod
    def comment_convert(cls, s, s_2, client=None):
        type = CommentType.from_value(s[i.COMMENT_TYPE])

        color_string = s.get(i.COMMENT_COLOR, '255,255,255')
        color = Colour.from_rgb(*map(int, color_string.split(',')))

        return Comment(
            body = b64.b64decode(mapper_util.normalize(s[i.COMMENT_BODY])).decode(),
            rating = s[i.COMMENT_RATING],
            timestamp = s[i.COMMENT_TIMESTAMP],
            id = s[i.COMMENT_ID],
            is_spam = bool(s.get(i.COMMENT_IS_SPAM, 0)),
            type = type,
            color = color,
            level_id = s.get(i.COMMENT_LEVEL_ID, 0),
            level_percentage = s.get(i.COMMENT_LEVEL_PERCENTAGE, -1),
            author = ClassConverter.abstractuser_convert(s_2, client),
            client = client
        )

    @classmethod
    def abstractuser_convert(cls, d, client=None):
        from_dict = {k: convert_to_type(v, int) for k, v in d.items()}
        return AbstractUser(**from_dict).attach_client(client)



# This is pretty much a copy of discord.py implementation.

import asyncio
import aiohttp
import gd
import inspect
import logging
import random
import time

MAX_ASYNCIO_SECONDS = 3456000

log = logging.getLogger(__name__)

class ExponentialBackoff:
    """An implementation of the exponential backoff algorithm.

    Provides a convenient interface to implement an exponential backoff
    for reconnecting or retrying transmissions in a distributed network.

    Once instantiated, the delay method will return the next interval to
    wait for when retrying a connection or transmission.  The maximum
    delay increases exponentially with each retry up to a maximum of
    2^10 * base, and is reset if no more attempts are needed in a period
    of 2^11 * base seconds.

    Parameters
    ----------
    base: :class:`int`
        The base delay in seconds. The first retry-delay will be up to
        this many seconds.
    integral: :class:`bool`
        Set to ``True`` if whole periods of base is desirable, otherwise any
        number in between may be returned.
    """

    def __init__(self, base=1, *, integral=False):
        self._base = base

        self._exp = 0
        self._max = 10
        self._reset_time = base * 2 ** 11
        self._last_invocation = time.monotonic()

        # Use our own random instance to avoid messing with global one
        rand = random.Random()
        rand.seed()

        self._randfunc = rand.randrange if integral else rand.uniform

    def delay(self):
        """Compute the next delay

        Returns the next delay to wait according to the exponential
        backoff algorithm.  This is a value between 0 and base * 2^exp
        where exponent starts off at 1 and is incremented at every
        invocation of this method up to a maximum of 10.

        If a period of more than base * 2^11 has passed since the last
        retry, the exponent is reset to 1.
        """
        invocation = time.monotonic()
        interval = invocation - self._last_invocation
        self._last_invocation = invocation

        if interval > self._reset_time:
            self._exp = 0

        self._exp = min(self._exp + 1, self._max)
        return self._randfunc(0, self._base * 2 ** self._exp)


class Loop:
    """A background task helper that abstracts the loop and reconnection logic for you.

    The main interface to create this is through :func:`loop`.
    """
    def __init__(self, coro, seconds, hours, minutes, count, reconnect, loop):
        self.coro = coro
        self.reconnect = reconnect
        self.loop = loop or asyncio.get_event_loop()
        self.count = count
        self._current_loop = 0
        self._task = None
        self._injected = None
        self._valid_exception = (
            OSError,
            gd.ClientException,
            aiohttp.ClientError,
            asyncio.TimeoutError,
        )

        self._before_loop = None
        self._after_loop = None
        self._is_being_cancelled = False
        self._has_failed = False
        self._stop_next_iteration = False

        if self.count is not None and self.count <= 0:
            raise ValueError('Count must be greater than 0 or None.')

        self.change_interval(seconds=seconds, minutes=minutes, hours=hours)

        if not inspect.iscoroutinefunction(self.coro):
            raise TypeError('Expected coroutine function, not {!r}.'.format(type(self.coro).__name__))

    async def _call_loop_function(self, name):
        coro = getattr(self, '_' + name)
        if coro is None:
            return

        if self._injected is not None:
            await coro(self._injected)
        else:
            await coro()

    async def _loop(self, *args, **kwargs):
        backoff = ExponentialBackoff()
        await self._call_loop_function('before_loop')
        try:
            while True:
                try:
                    await self.coro(*args, **kwargs)
                except self._valid_exception:
                    if not self.reconnect:
                        raise
                    await asyncio.sleep(backoff.delay())
                else:
                    if self._stop_next_iteration:
                        return
                    self._current_loop += 1
                    if self._current_loop == self.count:
                        break

                    await asyncio.sleep(self._sleep)
        except asyncio.CancelledError:
            self._is_being_cancelled = True
            raise
        except Exception:
            self._has_failed = True
            log.exception('Internal background task failed.')
            raise
        finally:
            await self._call_loop_function('after_loop')
            self._is_being_cancelled = False
            self._current_loop = 0
            self._stop_next_iteration = False
            self._has_failed = False

    def __get__(self, obj, objtype):
        if obj is None:
            return self
        self._injected = obj
        return self

    @property
    def current_loop(self):
        """:class:`int`: The current iteration of the loop."""
        return self._current_loop

    def start(self, *args, **kwargs):
        r"""Starts the internal task in the event loop.

        Parameters
        ------------
        \*args
            The arguments to to use.
        \*\*kwargs
            The keyword arguments to use.

        Raises
        --------
        RuntimeError
            A task has already been launched and is running.

        Returns
        ---------
        :class:`asyncio.Task`
            The task that has been created.
        """

        if self._task is not None and not self._task.done():
            raise RuntimeError('Task is already launched and is not completed.')

        if self._injected is not None:
            args = (self._injected, *args)

        self._task = self.loop.create_task(self._loop(*args, **kwargs))
        return self._task

    def stop(self):
        r"""Gracefully stops the task from running.

        Unlike :meth:`cancel`\, this allows the task to finish its
        current iteration before gracefully exiting.

        .. note::

            If the internal function raises an error that can be
            handled before finishing then it will retry until
            it succeeds.

            If this is undesirable, either remove the error handling
            before stopping via :meth:`clear_exception_types` or
            use :meth:`cancel` instead.
        """
        if self._task and not self._task.done():
            self._stop_next_iteration = True

    def _can_be_cancelled(self):
        return not self._is_being_cancelled and self._task and not self._task.done()

    def cancel(self):
        """Cancels the internal task, if it is running."""
        if self._can_be_cancelled():
            self._task.cancel()

    def restart(self, *args, **kwargs):
        r"""A convenience method to restart the internal task.

        .. note::

            Due to the way this function works, the task is not
            returned like :meth:`start`.

        Parameters
        ------------
        \*args
            The arguments to to use.
        \*\*kwargs
            The keyword arguments to use.
        """

        def restart_when_over(fut, *, args=args, kwargs=kwargs):
            self._task.remove_done_callback(restart_when_over)
            self.start(*args, **kwargs)

        if self._can_be_cancelled():
            self._task.add_done_callback(restart_when_over)
            self._task.cancel()

    def add_exception_type(self, exc):
        r"""Adds an exception type to be handled during the reconnect logic.

        This function is useful if you're interacting with a 3rd party library that
        raises its own set of exceptions.

        Parameters
        ------------
        exc: Type[:class:`BaseException`]
            The exception class to handle.

        Raises
        --------
        TypeError
            The exception passed is either not a class or not inherited from :class:`BaseException`.
        """
        if not inspect.isclass(exc):
            raise TypeError('{0!r} must be a class.'.format(exc))
        if not issubclass(exc, BaseException):
            raise TypeError('{0!r} must inherit from BaseException.'.format(exc))

        self._valid_exception = (*self._valid_exception, exc)

    def clear_exception_types(self):
        """Removes all exception types that are handled.

        .. note::

            This operation obviously cannot be undone!
        """
        self._valid_exception = tuple()

    def remove_exception_type(self, exc):
        """Removes an exception type from being handled during the reconnect logic.

        Parameters
        ------------
        exc: Type[:class:`BaseException`]
            The exception class to handle.

        Returns
        ---------
        :class:`bool`
            Whether it was successfully removed.
        """
        old_length = len(self._valid_exception)
        self._valid_exception = tuple(x for x in self._valid_exception if x is not exc)
        return len(self._valid_exception) != old_length

    def get_task(self):
        """Optional[:class:`asyncio.Task`]: Fetches the internal task or ``None`` if there isn't one running."""
        return self._task

    def is_being_cancelled(self):
        """Whether the task is being cancelled."""
        return self._is_being_cancelled

    def failed(self):
        """:class:`bool`: Whether the internal task has failed.

        .. versionadded:: 1.2.0
        """
        return self._has_failed

    def before_loop(self, coro):
        """A decorator that registers a coroutine to be called before the loop starts running.

        The coroutine must take no arguments (except ``self`` in a class context).

        Parameters
        ------------
        coro: :ref:`coroutine <coroutine>`
            The coroutine to register before the loop runs.

        Raises
        -------
        TypeError
            The function was not a coroutine.
        """

        if not inspect.iscoroutinefunction(coro):
            raise TypeError('Expected coroutine function, received {0.__name__!r}.'.format(type(coro)))

        self._before_loop = coro
        return coro

    def after_loop(self, coro):
        """A decorator that register a coroutine to be called after the loop finished running.

        The coroutine must take no arguments (except ``self`` in a class context).

        .. note::

            This coroutine is called even during cancellation. If it is desirable
            to tell apart whether something was cancelled or not, check to see
            whether :meth:`is_being_cancelled` is ``True`` or not.

        Parameters
        ------------
        coro: :ref:`coroutine <coroutine>`
            The coroutine to register after the loop finishes.

        Raises
        -------
        TypeError
            The function was not a coroutine.
        """

        if not inspect.iscoroutinefunction(coro):
            raise TypeError('Expected coroutine function, received {0.__name__!r}.'.format(type(coro)))

        self._after_loop = coro
        return coro

    def change_interval(self, *, seconds=0, minutes=0, hours=0):
        """Changes the interval for the sleep time.

        .. note::

            This only applies on the next loop iteration. If it is desirable for the change of interval
            to be applied right away, cancel the task with :meth:`cancel`.

        .. versionadded:: 1.2.0

        Parameters
        ------------
        seconds: :class:`float`
            The number of seconds between every iteration.
        minutes: :class:`float`
            The number of minutes between every iteration.
        hours: :class:`float`
            The number of hours between every iteration.

        Raises
        -------
        ValueError
            An invalid value was given.
        """

        sleep = seconds + (minutes * 60.0) + (hours * 3600.0)
        if sleep >= MAX_ASYNCIO_SECONDS:
            fmt = 'Total number of seconds exceeds asyncio imposed limit of {0} seconds.'
            raise ValueError(fmt.format(MAX_ASYNCIO_SECONDS))

        if sleep < 0:
            raise ValueError('Total number of seconds cannot be less than zero.')

        self._sleep = sleep
        self.seconds = seconds
        self.hours = hours
        self.minutes = minutes

def loop(*, seconds=0, minutes=0, hours=0, count=None, reconnect=True, loop=None):
    """A decorator that schedules a task in the background for you with
    optional reconnect logic.

    Parameters
    ------------
    seconds: :class:`float`
        The number of seconds between every iteration.
    minutes: :class:`float`
        The number of minutes between every iteration.
    hours: :class:`float`
        The number of hours between every iteration.
    count: Optional[:class:`int`]
        The number of loops to do, ``None`` if it should be an
        infinite loop.
    reconnect: :class:`bool`
        Whether to handle errors and restart the task
        using an exponential back-off algorithm.
    loop: :class:`asyncio.AbstractEventLoop`
        The loop to use to register the task, if not given
        defaults to :func:`asyncio.get_event_loop`.

    Raises
    --------
    :exc:`ValueError`
        An invalid value was given.

    :exc:`TypeError`
        The function was not a coroutine.

    Returns
    ---------
    :class:`Loop`
        The loop helper that handles the background task.
    """
    def decorator(func):
        return Loop(
            func, seconds=seconds, minutes=minutes, hours=hours,
            count=count, reconnect=reconnect, loop=loop)

    return decorator



import base64
import random

from typing import Union

from .mapper import mapper_util
from .crypto.coders import Coder

class Parameters:
    """Class that allows step-by-step parameter creation.
    Used in almost all HTTP Requests.

    A common usage of this class looks like that:

    .. code-block:: python3

        parameters = (
            Parameters().create_new()
            .put_definer('song', str(...))
            .finish()
        )
        await http.request(..., parameters)

        ...

    Attributes
    ----------
    gameVersion: :class:`str`
        Current game version.
    binaryVersion: :class:`str`
        Current binary version.
    secret: :class:`str`
        Secret used in requests.
    level_secret: :class:`str`
        Secret used in level operations (e.g. :meth:`.Level.delete`)
    login_secret: :class:`str`
        Secret used in login requests.
    mod_secret: :class:`str`
        Secret used in moderator requests.
    dict: :class:`dict`
        Dictionary where parameters are stored.
    """

    def __init__(self):
        self.gameVersion = '21'
        self.binaryVersion = '35'
        self.secret = 'Wmfd2893gb7'
        self.level_secret = 'Wmfv2898gc9'
        self.login_secret = 'Wmfv3899gc9'
        self.mod_secret = 'Wmfp3879gc3'
        self.dict = {}

    def create_new(self, type: str = None):
        """Start forming a new dictionary.

        Parameters
        ----------
        type: :class:`str`
            Type of start, use `'web'` to start with an empty dictionary.
            When omitted or ``None``, adds *gameVersion*, *binaryVersion*
            and *gdw* to a dict.

        Returns
        -------
        :class:`.Parameters`
            ``self``
        """
        if type is None:
            self.dict = {
                'gameVersion': self.gameVersion,
                'binaryVersion': self.binaryVersion,
                'gdw': '0'
            }
        if type == 'web':
            self.dict = {}
        return self

    def finish(self):
        """Finishes creating parameters dictionary, and adds ``secret`` parameter.

        Puts :attr:`.Parameters.secret` as *secret* parameter.

        Returns
        -------
        :class:`dict`
            Fully formatted dictionary.
        """
        self.dict['secret'] = self.secret
        return self.dict

    def finish_level(self):
        """Same as :meth:`.Parameters.finish`, but adds :attr:`.Parameters.level_secret` instead.

        Returns
        -------
        :class:`dict`
            Fully formatted dictionary.
        """
        self.dict['secret'] = self.level_secret
        return self.dict

    def finish_login(self):
        """Same as :meth:`.Parameters.finish`, but adds :attr:`.Parameters.login_secret` instead.

        Returns
        -------
        :class:`dict`
            Fully formatted dictionary.
        """
        self.dict['secret'] = self.login_secret
        return self.dict

    def finish_mod(self):
        """Same as :meth:`.Parameters.finish`, but adds :attr:`.Parameters.mod_secret` instead.

        Returns
        -------
        :class:`dict`
            Fully formatted dictionary.
        """
        self.dict['secret'] = self.mod_secret
        return self.dict

    def close(self):
        """Finishes creating parameters dictionary, without adding anything.

        Returns
        -------
        :class:`dict`
            Fully formatted parameters dictionary.
        """
        return self.dict

    def put_for_management(self, login: str, password: str, code: int):
        """Puts parameters for account management.

        Parameters
        ----------
        login: :class:`str`
            A username of an account.
        password: :class:`str`
            A password of an account.
        code: :class:`int`
            A solved RobTop's Captcha code.

        Returns
        -------
        :class:`.Parameters`
            ``self``
        """
        to_put = {
            'username': login,
            'password': password,
            'vercode': str(code),
            'cmdlogin': 'Login'
        }
        self.dict.update(to_put)
        return self

    def put_for_username(self, name: str, newname: str):
        """Adds parameters for username management.

        Parameters
        ----------
        name: :class:`str`
            An old username of an account.
        newname: :class:`str`
            A new username of an account.

        Returns
        -------
        :class:`.Parameters`
            ``self``
        """
        to_put = {
            'username': name,
            'newusername': newname,
            'changeusername': 'Change Username'
        }
        self.dict.update(to_put)
        return self

    def put_for_password(self, name: str, password: str, newpass: str):
        """Puts parameters for password management.

        Parameters
        ----------
        name: :class:`str`
            A username of an account.
        password: :class:`str`
            An old password of an account.
        newpass: :class:`str`
            A new password of an account.

        Returns
        -------
        :class:`.Parameters`
            ``self``
        """
        to_put = {
            'username': name,
            'oldpassword': password,
            'password': newpass,
            'password2': newpass,
            'change': 'Change Password'
        }
        self.dict.update(to_put)
        return self

    def put_definer(self, for_what: str, item):
        """Puts a definer.

        Parameters
        ----------
        for_what: :class:`str`
            Type of parameter to add, represented as string.

            Follows this mapping:

            .. code-block:: python3

                'song'      -> 'songID'
                'user'      -> 'targetAccountID'
                'search'    -> 'str'
                'password'  -> 'password'
                'levelid' -> 'levelID'
                'accountid' -> 'accountID'
                'itemid'    -> 'itemID'
                'messageid' -> 'messageID'
                'commentid' -> 'commentID'
                'requestid' -> 'requestID'
                'userid'    -> 'userID'
                'stars'     -> 'stars'
                'rating'    -> 'rating'

        item: `Any`
            Parameter to put.

        Returns
        -------
        :class:`.Parameters`
            ``self``
        """
        for_what = for_what.lower()
        params_dict = {
            'song': 'songID',
            'user': 'targetAccountID',
            'search': 'str',
            'password': 'password',
            'levelid': 'levelID',
            'accountid': 'accountID',
            'itemid': 'itemID',
            'messageid': 'messageID',
            'commentid': 'commentID',
            'requestid': 'requestID',
            'userid': 'userID',
            'stars': 'stars',
            'rating': 'rating'
        }
        try:
            self.dict[params_dict[for_what]] = str(item)
        except KeyError:
            pass
        return self

    def put_recipient(self, account_id: int):
        """Puts recipient of a message/friend request/...

        Parameters
        ----------
        account_id: :class:`int`
            An account ID of the recipient.

        Returns
        -------
        :class:`.Parameters`
            ``self``
        """
        self.dict['toAccountID'] = str(account_id)
        return self

    def put_is_sender(self, t: str):
        """Puts `'isSender'` parameter.

        Parameters
        ----------
        t: :class:`str`
            Either ``'normal'`` or ``'sent'``.
            If ``'sent'``, adds ``isSender=1``.

        Returns
        -------
        :class:`.Parameters`
            ``self``
        """
        if t == 'sent':
            self.dict['isSender'] = '1'

        return self

    def put_message(self, subject: str, body: str):
        """Puts message's subject and body.

        Parameters
        ----------
        subject: :class:`str`
            NOT ENCODED message subject.
        body: :class:`str`
            NOT ENCODED message body.

        Returns
        -------
        :class:`.Parameters`
            ``self``
        """
        self.dict['subject'] = base64.b64encode(subject.encode()).decode()
        self.dict['body'] = mapper_util.prepare_sending(Coder.encode(type='message', string=body))
        return self

    def put_password(self, item: str):
        """Self explanatory. Puts `'gjp'` parameter.

        Parameters
        ----------
        item: :class:`str`
            A password to put.

        Returns
        -------
        :class:`.Parameters`
            ``self``
        """
        self.dict['gjp'] = item
        return self

    def put_username(self, item: str):
        """Self explanatory. Puts `'userName'` parameter.

        Parameters
        ----------
        item: :class:`str`
            A username to put.

        Returns
        -------
        :class:`.Parameters`
            ``self``
        """
        self.dict['userName'] = item
        return self

    def put_fr_comment(self, item: str):
        """Pretty self explanatory. Puts `'comment'` parameter.

        Parameters
        ----------
        item: :class:`str`
            A comment to put.

        Returns
        -------
        :class:`.Parameters`
            ``self``
        """
        self.dict['comment'] = mapper_util.prepare_sending(
            base64.b64encode(item.encode()).decode())
        return self

    def put_type(self, number: Union[int, str] = 0):
        """Sets `'type'` parameter to a given number or string.

        Parameters
        ----------
        number: Union[:class:`int`, :class:`str`]
            A number or string to set type to.

        Returns
        -------
        :class:`.Parameters`
            ``self``
        """
        self.dict['type'] = str(number)
        return self

    def put_percent(self, number: int = 0):
        """Same as :meth:`.Parameters.put_type`, but for `'percent'`."""
        self.dict['percent'] = str(number)
        return self

    def put_page(self, number: int = 0):
        """Same as :meth:`.Parameters.put_type`, but for `'page'`."""
        self.dict['page'] = str(number)
        return self

    def put_weekly(self, number: int = 0):
        """Same as :meth:`.Parameters.put_type`, but for `'weekly'`."""
        self.dict['weekly'] = str(number)
        return self

    def put_total(self, number: int = 0):
        """Same as :meth:`.Parameters.put_type`, but for `'total'`."""
        self.dict['total'] = str(number)
        return self

    def put_mode(self, number: int = 0):
        """Same as :meth:`.Parameters.put_type`, but for `'mode'`."""
        self.dict['mode'] = str(number)
        return self

    def put_like(self, number: int = 0):
        """Same as :meth:`.Parameters.put_type`, but for `'like'`."""
        self.dict['like'] = str(number)
        return self

    def put_count(self, number: int = 0):
        """Same as :meth:`.Parameters.put_type`, but for `'count'`."""
        self.dict['count'] = str(number)
        return self

    def put_feature(self, number: int = 0):
        """Same as :meth:`.Parameters.put_type`, but for `'feature'`."""
        self.dict['feature'] = str(number)
        return self

    def put_special(self, number: int = 0):
        """Same as :meth:`.Parameters.put_type`, but for `'special'`."""
        self.dict['special'] = str(number)
        return self

    def put_local(self, number: int = 0):
        """Same as :meth:`.Parameters.put_type`, but for `'local'`."""
        self.dict['local'] = str(number)
        return self

    def put_rs(self, rs: str):
        """Self explanatory. Puts ``'rs'`` parameter.

        Parameters
        ----------
        rs: :class:`str`
            Random Seed parameter, as string.

        Returns
        -------
        :class:`.Parameters`
            ``self``
        """
        self.dict['rs'] = str(rs)
        return self

    def put_chk(self, chk: str):
        """Self explanatory. Puts ``'chk'`` parameter.

        Parameters
        ----------
        chk: :class:`str`
            Check parameter, as string.

        Returns
        -------
        :class:`.Parameters`
            ``self``
        """
        self.dict['chk'] = str(chk)
        return self

    def put_comment(self, content: str, values: list):
        """Puts a comment.

        Parameters
        ----------
        content: :class:`str`
            The content of the comment.
        values: :class:`list`
            A list of values to generate a ``chk`` parameter with.

        Returns
        -------
        :class:`.Parameters`
            ``self``
        """
        comment = mapper_util.prepare_sending(base64.b64encode(content.encode()).decode())
        self.dict['comment'] = comment
        values.insert(1, comment)
        self.put_chk(Coder.gen_chk(type='comment', values=values))
        return self

    def comment_for(self, type: str, number: int = None):
        """Defines type of a comment, and puts parameters required for it.

        Parameters
        ----------
        type: :class:`str`
            A type of the comment, either ``'profile'`` or ``'level'``.
            ``'profile'`` sets ``cType=1``, while ``'level'`` sets ``levelID=str(number)``.
        number: :class:`int`
            A number to put. Used as mentioned above.

        Returns
        -------
        :class:`.Parameters`
            ``self``
        """
        if type == 'profile':
            self.dict['cType'] = '1'

        elif type == 'level':
            self.dict['levelID'] = str(number)

        return self

    def put_login_definer(self, username: str, password: str):
        """Puts parameters for a login request.

        Parameters
        ----------
        username: :class:`str`
            Runs :meth:`.Parameters.put_username` with it.
        password: :class:`str`
            Adds this as a `'password'` parameter.

        Returns
        -------
        :class:`.Parameters`
            ``self``
        """
        self.dict.pop('gdw')  # it is not needed in login request
        self.dict['password'] = password
        self.put_udid()
        self.put_username(username)
        return self

    def put_udid(self, id: int = -1):
        """Puts ``'udid'`` parameter.

        .. note::

            If ``id`` is ``-1`` or omitted, random udid is generated.
            (From 100,000 to 100,000,000,000)

        Parameters
        ----------
        id: :class:`int`
            UDID to put.

        Returns
        -------
        :class:`.Parameters`
            ``self``
        """
        if not (id + 1):  # if -1
            id = random.randint(100_000, 100_000_000_000)

        self.dict['udid'] = str(id)
        return self


    def put_uuid(self, id: int = -1):
        """Same as :meth:`.Parameters.put_udid`, but puts ``'uuid'``.

        .. note::

            If ``id`` is ``-1``, generates randomly between 100,000 and 100,000,000.

        Parameters
        ----------
        id: :class:`int`
            UUID to put.

        Returns
        -------
        :class:`.Parameters`
            ``self``
        """
        if not (id + 1):
            id = random.randint(100_000, 100_000_000)

        self.dict['uuid'] = str(id)
        return self

    def put_level_desc(self, content: str):
        """Encodes given content and puts ``'levelDesc'``.

        Parameters
        ----------
        content: :class:`str`
            Content of the new description, NOT encoded in Base64.

        Returns
        -------
        :class:`.Parameters`
            ``self``
        """
        if content is None:
            content = ''

        desc = mapper_util.prepare_sending(base64.b64encode(content.encode()).decode())

        self.dict['levelDesc'] = desc
        return self

    def put_profile_upd(
        self, msg: int, friend_req: int, comments: int,
        youtube: str, twitter: str, twitch: str
    ):
        """Puts all parameters required for profile update.

        Parameters
        ----------
        msg: :class:`int`
            Message indicator. Mapped in as `'mS'`.
        friend_req: :class:`int`
            Friend request indicator. Mapped in as `'frS'`.
        msg: :class:`int`
            Comment history indicator. Mapped in as `'cS'`.
        youtube: :class:`str`
            Youtube username. Mapped in as `'yt'`.
        twitter: :class:`str`
            Twitter username. Mapped in as `'twitter'`.
        twitch: :class:`str`
            Twitch username. Mapped in as `'twitch'`.

        Returns
        -------
        :class:`.Parameters`
            ``self``
        """
        to_put = {
            'mS': str(msg), 'frS': str(friend_req), 'cS': str(comments),
            'yt': youtube, 'twitter': twitter, 'twitch': twitch
        }
        self.dict.update(to_put)
        return self

    def put_filters(self, filters: dict):
        """Appends level filters.

        Parameters
        ----------
        filters: Union[:class:`dict`, :class:`.Filters`]
            All filters as a dictionary or as a :class:`.Filters` object.

        Returns
        -------
        :class:`.Parameters`
            ``self``
        """
        if hasattr(filters, 'to_parameters'):
            filters = filters.to_parameters()
        self.dict.update(filters)
        return self

    def get_sent(self, indicator: int):
        if (indicator == 1):
            self.dict['getSent'] = '1'
        else:
            pass
        return self



import logging

from typing import Union

from .abstractentity import AbstractEntity
from .session import _session

from .errors import MissingAccess

from .utils.enums import DemonDifficulty, CommentStrategy, value_to_enum
from .utils.wrap_tools import make_repr, check

log = logging.getLogger(__name__)

class Level(AbstractEntity):
    """Class that represents a Geometry Dash Level.
    This class is derived from :class:`.AbstractEntity`.
    """
    def __init__(self, **options):
        super().__init__(**options)
        self._data = options.pop('data', '')
        self.options = options

    def __repr__(self):
        info = {
            'id': self.id,
            'name': self.name,
            'creator': self.creator,
            'version': self.version,
            'difficulty': self.difficulty
        }
        return make_repr(self, info)

    @property
    def name(self):
        """:class:`str`: The name of the level."""
        return self.options.get('name')

    @property
    def description(self):
        """:class:`str`: Description of the level."""
        return self.options.get('description')

    @property
    def version(self):
        """:class:`int`: Version of the level."""
        return self.options.get('version')

    @property
    def downloads(self):
        """:class:`int`: Amount of the level's downloads."""
        return self.options.get('downloads')

    @property
    def rating(self):
        """:class:`int`: Amount of the level's likes or dislikes."""
        return self.options.get('rating')

    @property
    def score(self):
        """:class:`int`: Level's featured score."""
        return self.options.get('score')

    @property
    def creator(self):
        """:class:`.AbstractUser`: Creator of the level."""
        return self.options.get('creator')

    @property
    def song(self):
        """:class:`.Song`: Song used in the level."""
        return self.options.get('song')

    @property
    def difficulty(self):
        """:class:`.LevelDifficulty`: Difficulty of the level."""
        return self.options.get('difficulty')

    @property
    def stars(self):
        """:class:`int`: Amount of stars the level has."""
        return self.options.get('stars')

    @property
    def coins(self):
        """:class:`int`: Amount of coins in the level."""
        return self.options.get('coins')

    @property
    def original_id(self):
        """:class:`int`: ID of the original level. (``0`` if is not a copy)"""
        return self.options.get('original')

    @property
    def uploaded_timestamp(self):
        """:class:`str`: A human-readable string representing how much time ago level was uploaded."""
        return self.options.get('uploaded_timestamp')

    @property
    def last_uploaded_timestamp(self):
        """:class:`str`: A human-readable string showing how much time ago the last update was."""
        return self.options.get('last_updated_timestamp')

    @property
    def length(self):
        """:class:`.LevelLength`: A type that represents length of the level."""
        return self.options.get('length')

    @property
    def game_version(self):
        """:class:`int`: A version of the game required to play the level."""
        return self.options.get('game_version')

    @property
    def requested_stars(self):
        """:class:`int`: Amount of stars creator of the level has requested."""
        return self.options.get('stars_requested')

    @property
    def object_count(self):
        """:class:`int`: Amount of objects the level has."""
        return self.options.get('object_count')

    @property
    def type(self):
        """:class:`.TimelyType`: A type that shows whether a level is Daily/Weekly."""
        return self.options.get('type')

    @property
    def timely_index(self):
        """:class:`int`: A number that represents current index of the timely.
        Increments on new dailies/weeklies. If not timely, equals ``-1``.
        """
        return self.options.get('time_n')

    @property
    def cooldown(self):
        """:class:`int`: Represents a cooldown until next timely. If not timely, equals ``-1``."""
        return self.options.get('cooldown')

    def is_timely(self, daily_or_weekly: str = None):
        """:class:`bool`: Indicates whether a level is timely/daily/weekly.
        For instance, let's suppose a *level* is daily. Then, the behavior of this method is:
        ``level.is_timely() -> True`` and ``level.is_timely('daily') -> True`` but
        ``level.is_timely('weekly') -> False``."""
        if self.type is None:
            return False

        if daily_or_weekly is None:
            return self.type.value > 0

        assert daily_or_weekly in ('daily', 'weekly')

        return self.type.name.lower() == daily_or_weekly

    def is_featured(self):
        """:class:`bool`: Indicates whether a level is featured."""
        return self.score > 0  # not sure if this is the right way though

    def is_epic(self):
        """:class:`bool`: Indicates whether a level is epic."""
        return self.options.get('is_epic')

    def is_demon(self):
        """:class:`bool`: Indicates whether a level is demon."""
        return options.get('is_demon')

    def is_auto(self):
        """:class:`bool`: Indicates whether a level is auto."""
        return options.get('is_auto')

    def is_original(self):
        """:class:`bool`: Indicates whether a level is original."""
        return not self.original_id

    def has_coins_verified(self):
        """:class:`bool`: Indicates whether level's coins are verified."""
        return self.options.get('verified_coins')

    def download(self):
        """:class:`str`: Returns level data, represented as string."""
        return self._data

    async def report(self):
        """|coro|

        Reports a level.

        Raises
        ------
        :exc:`.MissingAccess`
            Failed to report a level.
        """
        await _session.report_level(self)

    async def delete(self, *, from_client=None):
        """|coro|

        Deletes a level.

        Parameters
        ----------
        from_client: :class:`.Client`
            A logged in client to delete a level with. If ``None`` or omitted,
            defaults to the one attached to this level.

        Raises
        ------
        :exc:`.MissingAccess`
            Failed to delete a level.
        """
        client = from_client if from_client is not None else self._client
        check.is_logged_obj(client, 'delete')
        await _session.delete_level(self, client=client)

    async def update_description(self, content: str = None, *, from_client=None):
        """|coro|

        Updates level description.

        Parameters
        ----------
        content: :class:`str`
            Content of the new description. If ``None`` or omitted,
            sets content to :attr:`.Level.description`.

        from_client: :class:`.Client`
            A logged in client to update level description with. If ``None`` or omitted,
            defaults to the one attached to this level.

        Raises
        ------
        :exc:`.MissingAccess`
            Failed to update level's description.
        """
        if content is None:
            content = self.description

        client = from_client if from_client is not None else self._client
        check.is_logged_obj(client, 'update_description')
        await _session.update_level_desc(self, content, client=client)

    async def rate(self, stars: int = 1, *, from_client=None):
        """|coro|

        Sends level rating.

        Parameters
        ----------
        stars: :class:`int`
            Amount of stars to rate with.

        from_client: :class:`.Client`
            A logged in client to rate level with. If ``None`` or omitted,
            defaults to the one attached to this level.

        Raises
        ------
        :exc:`.MissingAccess`
            Failed to rate a level.
        """
        client = from_client if from_client is not None else self._client
        check.is_logged_obj(client, 'rate')
        await _session.rate_level(self, stars, client=client)

    async def rate_demon(
        self, demon_difficulty: Union[int, str, DemonDifficulty] = 1,
        as_mod: bool = False, *, from_client=None
    ):
        """|coro|

        Sends level demon rating.

        Parameters
        ----------
        demon_difficulty: Union[:class:`int`, :class:`str`, :class:`.DemonDifficulty`]
            Demon difficulty to rate a level with.

        as_mod: :class:`bool`
            Whether to send a demon rating as moderator.

        from_client: :class:`.Client`
            A logged in client to rate demon difficulty with. If ``None`` or omitted,
            defaults to the one attached to this level.

        Raises
        ------
        :exc:`.MissingAccess`
            If attempted to rate a level as moderator without required permissions.
        """
        client = from_client if from_client is not None else self._client
        check.is_logged_obj(client, 'rate_demon')
        demon_difficulty = value_to_enum(DemonDifficulty, demon_difficulty)

        success = await _session.rate_demon(self, demon_difficulty, mod=as_mod, client=client)

        if success:
            log.info('Successfully demon-rated level: %s.', self)
        else:
            log.warning('Failed to rate demon difficulty for level: %s.', self)

    async def send(self, stars: int = 1, featured: bool = True, *, from_client=None):
        """|coro|

        Sends a level to Geometry Dash Developer and Administrator, *RobTop*.

        Parameters
        ----------
        stars: :class:`int`
            Amount of stars to send with.

        featured: :class:`bool`
            Whether to send to feature, or to simply rate.

        from_client: :class:`.Client`
            A logged in client to send a level from. If ``None`` or omitted,
            defaults to the one attached to this level.

        Raises
        ------
        :exc:`.MissingAccess`
            Missing required moderator permissions.
        """
        client = from_client if from_client is not None else self._client
        check.is_logged_obj(client, 'send')
        await _session.send_level(self, stars, featured=featured, client=client)

    async def is_alive(self):
        """|coro|

        Checks if a level is still on Geometry Dash servers.

        Returns
        -------
        :class:`bool`
            ``True`` if a level is still *alive*, and ``False`` otherwise.
        """
        new_version = await self.refresh()
        return not (new_version is None)

    async def refresh(self):
        """|coro|

        Refreshes a level. Returns ``None`` on fail.

        .. note::

            This function actually refreshes a level and its stats.
            No need to do funky stuff with its return.

        Returns
        -------
        :class:`.Level`
            A newly fetched version. ``None`` if failed to fetch.
        """
        try:
            if self.is_timely():
                new_ver = await _session.get_timely(self.type.name.lower(), client=self._client)

                if new_ver.id != self.id:
                    log.warning('There is a new %s.type: %s. Updating to it...', self, new_ver)

            else:
                new_ver = await _session.get_level(self.id, client=self._client)

        except MissingAccess:
            return log.warning('Failed to refresh level: %r. Most likely it was deleted.', self)

        self.options = new_ver.options
        self._data = new_ver._data

        return self

    async def comment(self, content: str, percentage: int = 0, *, from_client=None):
        """|coro|

        Posts a comment on a level.

        Parameters
        ----------
        content: :class:`str`
            Body of the comment to post.

        percentage: :class:`int`
            Percentage to display. Default is ``0``.

            .. note::

                gd.py developers are not responsible for effects that changing this may cause.
                Set this parameter higher than 0 on your own risk.

        from_client: :class:`.Client`
            A logged in client to post a comment from. If ``None`` or omitted,
            defaults to the one attached to this level.

        Raises
        ------
        :exc:`.MissingAccess`
            Failed to post a level comment.
        """
        client = from_client if from_client is not None else self._client
        check.is_logged_obj(client, 'comment')
        await _session.comment_level(self, content, percentage, client=client)

    async def like(self, from_client=None):
        """|coro|

        Likes a level.

        Parameters
        ----------
        from_client: :class:`.Client`
            A logged in client to like a level with. If ``None`` or omitted,
            defaults to the one attached to this level.

        Raises
        ------
        :exc:`.MissingAccess`
            Failed to like a level.
        """
        client = from_client if from_client is not None else self._client
        check.is_logged_obj(client, 'like')
        await _session.like(self, dislike=False, client=client)

    async def dislike(self, from_client=None):
        """|coro|

        Dislikes a level.

        Parameters
        ----------
        from_client: :class:`.Client`
            A logged in client to dislike a level with. If ``None`` or omitted,
            defaults to the one attached to this level.

        Raises
        ------
        :exc:`.MissingAccess`
            Failed to dislike a level.
        """
        client = from_client if from_client is not None else self._client
        check.is_logged_obj(client, 'dislike')
        await _session.like(self, dislike=True, client=client)

    async def get_comments(self, strategy: Union[int, str, CommentStrategy] = 0, amount: int = 20):
        """|coro|

        Retrieves level comments.

        Parameters
        ----------
        strategy: Union[:class:`int`, :class:`str`, :class:`.CommentStrategy`]
            A strategy to apply when searching. This is converted to :class:`.CommentStrategy`
            using :func:`.utils.value_to_enum`.

        amount: :class:`int`
            Amount of comments to retrieve. Default is ``20``. (no limits)

        Returns
        -------
        List[:class:`.Comment`]
            List of comments retrieved.

        Raises
        ------
        :exc:`.MissingAccess`
            Failed to fetch comments.

        :exc:`.NothingFound`
            No comments were found.

        :exc:`.FailedConversion`
            Raised if ``strategy`` can not be converted to :class:`.CommentStrategy`.
        """
        strategy = value_to_enum(CommentStrategy, strategy)

        return await _session.get_level_comments(level=self, strategy=strategy, amount=amount)



from .utils.wrap_tools import make_repr

class AbstractEntity:
    """Class that represents Abstract Entity. This is a base for many gd.py objects.

    .. container:: operations

        .. describe:: x == y

            Check if two objects are equal. Compared by hash and type.

        .. describe:: x != y

            Check if two objects are not equal

        .. describe:: hash(x)

            Returns ``hash(self.to_hash_string())``.
    """
    def __init__(self, **options):
        self._id = options.get('id', 0)
        self._client = options.get('client')  # None if not provided

    def __repr__(self):
        info = {'id': self.id}
        return make_repr(self, info)

    def __hash__(self):
        return hash(self.to_hash_string())

    def __eq__(self, other):
        return isinstance(other, type(self)) and hash(self) == hash(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def id(self):
        """:class:`int`: ID of the Entity."""
        return self._id

    def to_hash_string(self):
        """:class:`str`: Converts to a string which is used in ``hash(self)``."""
        cls = self.__class__.__name__
        id = self._id
        r = self.__repr__().replace(' ', '_').lower()

        hash_string = '<GDEntity<{cls}>[id={id};repr={r}]>'.format(cls=cls, id=id, r=r)

        return hash_string

    def attach_client(self, client):
        self._client = client
        return self



from .abstractentity import AbstractEntity
from .session import _session

from .utils.wrap_tools import make_repr, check

class Message(AbstractEntity):
    """Class that represents private messages in Geometry Dash.
    This class is derived from :class:`.AbstractEntity`.
    """
    def __init__(self, **options):
        super().__init__(**options)
        self.options = options
        self._body = None

    def __repr__(self):
        info = {
            'author': self.author,
            'id': self.id,
            'is_read': self.is_read()
        }
        return make_repr(self, info)

    @property
    def author(self):
        """:class:`.AbstractUser`: An author of the message."""
        return self.options.get('author')

    @property
    def recipient(self):
        """:class:`.AbstractUser`: A recipient of the message."""
        return self.options.get('recipient')

    @property
    def subject(self):
        """:class:`str`: A subject of the message, as string."""
        return self.options.get('subject')

    @property
    def timestamp(self):
        """:class:`str`: A human-readable string representing how long ago message was created."""
        return self.options.get('timestamp')

    @property
    def type(self):
        """:class:`.MessageOrRequestType`: Whether a message is sent or inbox."""
        return self.options.get('type')

    @property
    def body(self):
        """Union[:class:`str`, ``None``]: A body of the message. Requires :meth:`.Message.read`."""
        return self._body

    def is_read(self):
        """:class:`bool`: Indicates whether message is read or not."""
        return self.options.get('is_read')

    @check.is_logged()
    async def read(self):
        """|coro|

        Read a message. Set the body of the message to the content.

        Returns
        -------
        :class:`str`
            The content of the message.
        """
        await _session.read_message(self)

    @check.is_logged()
    async def delete(self):
        """|coro|

        Delete a message.

        Raises
        ------
        :exc:`.MissingAccess`
            Failed to delete a message.
        """
        return await _session.delete_message(self)



from .abstractentity import AbstractEntity
from .session import _session

from .utils.wrap_tools import make_repr, check

class Comment(AbstractEntity):
    """Class that represents a Profile/Level comment in Geometry Dash.
    This class is derived from :class:`.AbstractEntity`.
    """
    def __init__(self, **options):
        super().__init__(**options)
        # self._client = client
        self.options = options

    def __repr__(self):
        info = {
            'author': self.author,
            'id': self.id,
            'rating': self.rating
        }
        return make_repr(self, info)

    @property
    def body(self):
        """:class:`str`: Returns body of the comment."""
        return self.options.get('body')

    @property
    def rating(self):
        """:class:`int`: Rating of the comment."""
        return self.options.get('rating')

    @property
    def timestamp(self):
        """:class:`str`: A human-readable timestamp representing how long ago the comment was created."""
        return self.options.get('timestamp')

    @property
    def author(self):
        """:class:`.AbstractUser`: An author of the comment."""
        return self.options.get('author')

    @property
    def type(self):
        """:class:`.CommentType`: Whether comment is on profile or on a level."""
        return self.options.get('type')

    @property
    def level_id(self):
        """:class:`int`: Level ID of a level the comment is on, if present. ``0`` if profile comment."""
        return self.options.get('level_id')

    @property
    def level_percentage(self):
        """:class:`int`: Level highscore linked to a comment, if present. ``-1`` if profile comment."""
        return self.options.get('level_percentage')

    @property
    def color(self):
        """:class:`.Colour`: Colour of the comment. Oftenly equals ``gd.Colour(0xffffff)``."""
        return self.options.get('color')

    def is_spam(self):
        """:class:`bool`: Indicates whether a comment is marked as spam. ``False`` if profile comment."""
        return self.options.get('is_spam')

    def is_disliked(self):
        """:class:`bool`: Indicates whether a comment is disliked or not."""
        return abs(self.rating) != self.rating

    async def like(self, from_client=None):
        """|coro|

        Likes a comment.

        Parameters
        ----------
        from_client: :class:`.Client`
            A logged in client to like with. If ``None`` or omitted,
            defaults to the one attached to this comment.

        Raises
        ------
        :exc:`.MissingAccess`
            Failed to like a comment.
        """
        client = from_client if from_client is not None else self._client
        check.is_logged_obj(client, 'like')
        await _session.like(self, dislike=False, client=client)

    async def dislike(self, from_client=None):
        """|coro|

        Dislikes a comment.

        Parameters
        ----------
        from_client: :class:`.Client`
            A logged in client to dislike comment with. If ``None`` or omitted,
            defaults to the one attached to this comment.

        Raises
        ------
        :exc:`.MissingAccess`
            Failed to dislike a comment.
        """
        client = from_client if from_client is not None else self._client
        check.is_logged_obj(client, 'dislike')
        await _session.like(self, dislike=True, client=client)

    async def delete(self, from_client=None):
        """|coro|

        Deletes a comment from Geometry Dash servers.

        Obviously, only comments of client logged in using :meth:`.Client.login` can be deleted.

        The normal behaviour of the server is returning 1 if comment was deleted,
        so the error is raised on response != 1.

        Parameters
        ----------
        from_client: :class:`.Client`
            A logged in client to delete comments of. If ``None`` or omitted,
            defaults to the one attached to this comment.

        Raises
        ------
        :exc:`.MissingAccess`
            Server did not return 1, which means comment was not deleted.
        """
        client = from_client if from_client is not None else self._client
        check.is_logged_obj(client, 'delete')
        await _session.delete_comment(self, client=client)



class GDException(Exception):
    """Base exception class for gd.py.
    This could be caught to handle any exceptions thrown from this library.
    """
    pass


class ClientException(GDException):
    """Base exception class for errors that are thrown
    when operation in :class:`Client` fails.
    """
    pass


class PaginatorException(GDException):
    """Base exception class for errors that are thrown
    when operating with :class:`Paginator` gets an error.
    """
    pass


class FailedConversion(GDException):
    """Exception that is raised when Enum converter
    fails to turn given value into requested Enum.
    """
    def __init__(self, enum, value):
        self._enum = enum
        self._value = value
        message = 'Failed to convert value {!r} to enum: {!r}.'.format(value, enum)
        super().__init__(message)

    @property
    def enum(self):
        """:class:`enum.Enum`: Enum to which ``value`` was failed to convert."""
        return self._enum

    @property
    def value(self):
        """`Any`: Value that was failed to be converted to ``enum``."""
        return self._value


class HTTPNotConnected(ClientException):
    """Exception that is raised when exception
    in :class:`.utils.http.HTTPClient` occurs.
    """
    def __init__(self):
        message = 'Internet connection failed.'
        super().__init__(message)


class FailedCaptcha(ClientException):
    """Exception that is raised when unknown results
    when solving GD Captcha are recieved.
    """
    def __init__(self, msg):
        message = 'Solving Captcha failed. {}'.format(msg)
        super().__init__(message)


class MissingAccess(ClientException):
    """Exception that is raised when server responses with -1."""
    def __init__(self, *, message: str = None):
        super().__init__(message)


class SongRestrictedForUsage(ClientException):
    """Exception that is raised when server returns -2
    when looking for a song.
    """
    def __init__(self, id):
        message = 'Song with id {!r} is not allowed to use.'.format(id)
        super().__init__(message)


class LoginFailure(ClientException):
    """Exception that is raised when server returns -1
    when trying to log in.
    """
    def __init__(self, login, password):
        self._login = login
        self._password = password

        message = (
            'Failed to login with parameters: '
            '<login={0.login!r}, password={0.password!r}>.'
        ).format(self)

        super().__init__(message)

    @property
    def login(self):
        """Username that was wrong or password did not match."""
        return self._login

    @property
    def password(self):
        """Password that login was failed with."""
        return self._password


class FailedToChange(ClientException):
    """Exception that is raised when logged in :class:`Client`
    fails to change its password or username.
    """
    def __init__(self, type):
        message = 'Failed to change {}. Reason: Unspecified'.format(type)  # [Future]
        super().__init__(message)


class NothingFound(ClientException):
    """Exception that is raised when server returns nothing
    that can be converted to object of name *cls_name*.
    """
    def __init__(self, cls_name):
        self._cls_name = cls_name
        message = 'No <{}> instances were found.'.format(self.cls_name)
        super().__init__(message)

    @property
    def cls_name(self):
        """Name of the class instances of which were not found."""
        return self._cls_name


class NotLoggedError(ClientException):
    """Exception that is raised when a function that requires logged in user is called
    while :class:`Client` is not logged.
    """
    def __init__(self, func_name):
        message = '{!r} requires client to be logged.'.format(func_name)
        super().__init__(message)


class PagesOutOfRange(PaginatorException):
    """Exception that is raised if a non-existing page
    is requested in :class:`Paginator`.
    """
    def __init__(self, page, info):

        if str(info).isdigit():
            message = 'Pages are out of range. Requested page: {!r}, Pages existing: {}'.format(page, info)
        else:
            message = '{} Requested page: {}'.format(info, page)

        super().__init__(message)



from .abstractentity import AbstractEntity
from .session import _session

from .utils.wrap_tools import make_repr, check

class FriendRequest(AbstractEntity):
    """Class that represents a friend request.
    This class is derived from :class:`.AbstractEntity`.
    """
    def __init__(self, **options):
        super().__init__(**options)
        self.options = options

    def __repr__(self):
        info = {
            'id': self.id,
            'author': self.author,
            'type': self.type
        }
        return make_repr(self, info)

    @property
    def author(self):
        """:class:`.AbstractUser`: An author of the friend request."""
        return self.options.get('author')

    @property
    def recipient(self):
        """:class:`.AbstractUser`: A recipient of the friend request."""
        return self.options.get('recipient')

    @property
    def type(self):
        """:class:`.MessageOrRequestType`: Whether request is incoming or sent."""
        return self.options.get('type')

    @property
    def body(self):
        """:class:`str`: A string representing a message request was sent with."""
        return self.options.get('body')

    @property
    def timestamp(self):
        """:class:`str`: A human-readable string representing how long ago request was created."""
        return self.options.get('timestamp')

    def is_read(self):
        """:class:`bool`: Indicates whether request was already read."""
        return self.options.get('is_read')

    @check.is_logged()
    async def read(self):
        """|coro|

        Read a friend request. Sets ``is_read`` to ``True`` on success.

        Raises
        ------
        :exc:`.MissingAccess`
            Failed to read a message.
        """
        await _session.read_friend_req(self)

    @check.is_logged()
    async def delete(self):
        """|coro|

        Delete a friend request.

        Raises
        ------
        :exc:`.MissingAccess`
            Failed to delete a request.
        """
        await _session.delete_friend_req(self)

    @check.is_logged()
    async def accept(self):
        """|coro|

        Accept a friend request.

        Raises
        ------
        :exc:`.MissingAccess`
            Failed to accept a request.
        """
        await _session.accept_friend_req(self)



from .utils.wrap_tools import make_repr

class IconSet:
    """Class that represents an Icon Set."""
    def __init__(self, **options):
        self.options = options

    def __repr__(self):
        info = {
            'main_icon': self.main,
            'main_type': self.main_type,
            'color_1': self.color_1,
            'color_2': self.color_2
        }
        return make_repr(self, info)

    @property
    def main(self):
        """:class:`int`: ID of the main icon of the iconset. (see :attr:`.IconSet.main_type`)"""
        return self.options.get('main_icon')

    @property
    def color_1(self):
        """:class:`.Colour`: A first color of the iconset."""
        return self.options.get('color_1')

    @property
    def color_2(self):
        """:class:`.Colour`: A second color of the iconset."""
        return self.options.get('color_2')

    @property
    def main_type(self):
        """:class:`.IconType`: A type of the main icon of the iconset."""
        return self.options.get('main_icon_type')

    @property
    def cube(self):
        """:class:`int`: Cube ID of the iconset."""
        return self.options.get('icon_cube')

    @property
    def ship(self):
        """:class:`int`: Ship ID of the iconset."""
        return self.options.get('icon_ship')

    @property
    def ball(self):
        """:class:`int`: Ball ID of the iconset."""
        return self.options.get('icon_ball')

    @property
    def ufo(self):
        """:class:`int`: UFO ID of the iconset."""
        return self.options.get('icon_ufo')

    @property
    def wave(self):
        """:class:`int`: Wave ID of the iconset."""
        return self.options.get('icon_wave')

    @property
    def robot(self):
        """:class:`int`: Robot ID of the iconset."""
        return self.options.get('icon_robot')

    @property
    def spider(self):
        """:class:`int`: Spider ID of the iconset."""
        return self.options.get('icon_spider')

    def has_glow_outline(self):
        """:class:`bool`: Indicates whether an iconset has glow outline."""
        return self.options.get('has_glow_outline')

    def get_colors(self):
        """Tuple[:class:`.Colour`, :class:`.Colour`]: A shorthand for *color_1* and *color_2*."""
        return self.color_1, self.color_2



import asyncio
import itertools
import re  # for NG songs
import time  # for perf_counter in ping

from typing import Union, Sequence, Tuple

from .unreguser import UnregisteredUser
from .errors import *

from .utils.captcha_solver import Captcha
from .utils.converter import Converter
from .utils.enums import SearchStrategy
from .utils.filters import Filters
from .utils.http_request import http
from .utils.indexer import Index as i
from .utils.mapper import mapper_util
from .utils.params import Parameters as Params
from .utils.routes import Route
from .utils.save_parser import SaveParser
from .utils.wrap_tools import check
from .utils.crypto.coders import Coder

from . import utils

class GDSession:
    """Implements all requests-related functionality.
    No docstrings here yet...
    I am sorry for all in-def imports.
    - NeKitDS
    """
    async def ping_server(self, link: str):
        start = time.perf_counter()
        await http.normal_request(link)
        end = time.perf_counter()

        ping = (end-start)*1000

        return round(ping, 2)


    async def get_song(self, song_id: int = 0):
        from .classconverter import ClassConverter

        parameters = Params().create_new().put_definer('song', song_id).finish()
        codes = {
            -1: MissingAccess(message='No songs were found with ID: {}.'.format(song_id)),
            -2: SongRestrictedForUsage(song_id)
        }
        resp = await http.request(
            Route.GET_SONG_INFO, parameters, splitter='~|~', error_codes=codes, should_map=True)
        return ClassConverter.song_convert(resp)


    async def get_ng_song(self, song_id: int = 0):
        # just like get_song(), but gets anything available on NG.
        from .classconverter import ClassConverter

        song_id = int(song_id)  # ensure type

        link = Route.NEWGROUNDS_SONG_LISTEN + str(song_id)
        resp = await http.normal_request(link)
        content = await resp.content.read()
        html = content.decode().replace('\\', '')

        RE = (
            'https:\/\/audio\.ngfiles\.com\/([^\'"]+)',  # searching for link
            '.filesize.:(\d+)',  # searching for size
            '<title>([^<>]+)<\/title>',  # searching for name
            '.artist.:.([^\'"]+).'  # searching for author
        )
        try:
            dl_link = re.search(RE[0], html).group(0)
            size_b = int(re.search(RE[1], html).group(1))  # in B
            size_mb = round(size_b/1024/1024, 2)  # in MB (rounded)
            name = re.search(RE[2], html).group(1)
            author = re.search(RE[3], html).group(1)
        except AttributeError:  # if re.search returned None -> Song not found
            raise MissingAccess(message='No song found under ID: {}.'.format(song_id))

        return ClassConverter.song_from_kwargs(
            name=name, author=author, id=song_id, size=size_mb, links=[link, dl_link], custom=True
        )


    async def get_user(self, account_id: int = 0, return_only_stats: bool = False, *, client):
        from .classconverter import ClassConverter

        if account_id == -1:
            return UnregisteredUser()

        parameters = Params().create_new().put_definer('user', account_id).finish()
        codes = {
            -1: MissingAccess(message='No users were found with ID: {}.'.format(account_id))
        }

        resp = await http.request(
            Route.GET_USER_INFO, parameters, splitter=':', error_codes=codes, should_map=True)

        if return_only_stats:
            return ClassConverter.user_stats_convert(resp, client)

        another = (
            Params().create_new().put_definer('search', resp.get(i.USER_PLAYER_ID))
            .put_total(0).put_page(0).finish()
        )
        some_resp = await http.request(Route.USER_SEARCH, another, splitter='#')

        item = some_resp[0]
        if not item:
            return

        new_resp = mapper_util.map(item.split(':'))
        to_add = (i.USER_NAME, i.USER_ICON, i.USER_ICON_TYPE)
        new_dict = {k: new_resp.get(k) for k in to_add}
        resp.update(new_dict)

        return ClassConverter.user_convert(resp, client)


    def to_user(self, conv_dict: dict = None, *, client):
        from .classconverter import ClassConverter

        if conv_dict is None:
            conv_dict = {}

        return ClassConverter.abstractuser_convert(conv_dict, client)


    async def search_user(self, param: Union[str, int] = None, return_abstract: bool = False, *, client):
        from .classconverter import ClassConverter

        assert param is not None

        parameters = (
            Params().create_new().put_definer('search', param)
            .put_total(0).put_page(0).finish()
        )
        codes = {
            -1: MissingAccess(message='Searching for {} returned -1.'.format(param))
        }

        resp = await http.request(Route.USER_SEARCH, parameters, splitter='#', error_codes=codes)
        item = resp[0]
        if not item:
            return

        mapped = mapper_util.map(item.split(':'))

        name = mapped.get(i.USER_NAME, '') or 'UnregisteredUser'
        id = mapped.get(i.USER_PLAYER_ID, 0) or 0
        account_id = mapped.get(i.USER_ACCOUNT_ID, 0) or 0

        if not id or not account_id:
            return UnregisteredUser(name=name, id=id)

        if return_abstract:
            ret = dict(name=name, id=id, account_id=account_id)
            return ClassConverter.abstractuser_convert(ret, client)

        # ok if we should not return abstract, let's find all other parameters
        parameters = Params().create_new().put_definer('user', mapped.get(i.USER_ACCOUNT_ID, 0)).finish()

        resp = await http.request(
            Route.GET_USER_INFO, parameters, splitter=':', error_codes=codes, should_map=True
        )

        resp.update(
            {k: mapped.get(k) for k in (i.USER_NAME, i.USER_ICON, i.USER_ICON_TYPE)}
        )

        return ClassConverter.user_convert(resp, client)


    async def get_level(
        self, level_id: int = 0, timetuple: Tuple[int, int, int] = (0, -1, -1), *, client
    ):
        from .classconverter import ClassConverter

        assert level_id >= -2

        type, number, cooldown = map(str, timetuple)
        ext = ['101', type, '102', number, '103', cooldown, '104', '0']

        codes = {
            -1: MissingAccess(message='Failed to get a level. Given ID: {}'.format(level_id))
        }

        parameters = Params().create_new().put_definer('levelid', level_id).finish()
        resp = await http.request(Route.DOWNLOAD_LEVEL, parameters, error_codes=codes, splitter='#')

        level_data = mapper_util.map(resp[0].split(':') + ext)

        real_id = level_data.get(i.LEVEL_ID)

        parameters = (
            Params().create_new().put_definer('search', real_id)
            .put_filters(Filters.setup_empty()).finish()
        )
        resp = await http.request(Route.LEVEL_SEARCH, parameters, error_codes=codes, splitter='#')

        # getting song
        song_data = resp[2]
        song = (
            Converter.to_normal_song(level_data.get(i.LEVEL_AUDIO_TRACK)) if not song_data
            else ClassConverter.song_convert(mapper_util.map(song_data.split('~|~')))
        ).attach_client(client)

        # getting creator
        creator_data = resp[1]
        creator = (
            UnregisteredUser(level_data.get(i.LEVEL_CREATOR_ID)) if not creator_data
            else ClassConverter.abstractuser_convert(
                {k: v for k, v in zip(('id', 'name', 'account_id'), creator_data.split(':'))}
            )
        ).attach_client(client)

        return ClassConverter.level_convert(
            level_data, song=song, creator=creator, client=client)


    async def get_timely(self, type: str = 'daily', *, client):
        w = ('daily', 'weekly').index(type)
        parameters = Params().create_new().put_weekly(w).finish()
        codes = {
            -1: MissingAccess(message='Failed to fetch a {!r} level.'.format(type))
        }
        resp = await http.request(Route.GET_TIMELY, parameters, error_codes=codes, splitter='|')
        if len(resp) != 2:
            raise MissingAccess(message='Unknown response: {}.'.format(resp))

        num, cooldown = map(int, resp)
        num %= 100_000
        w += 1

        level = await self.get_level(-w, (w, num, cooldown), client=client)
        return level.attach_client(client)


    async def get_friends(self, client):
        from .classconverter import ClassConverter

        parameters = (
            Params().create_new().put_definer('accountid', client.account_id)
            .put_password(client.encodedpass).put_type(0).finish()
        )
        codes = {
            -1: MissingAccess(message='Failed to get friends.'),
            -2: NothingFound('gd.AbstractUser')
        }

        resp = await http.request(Route.GET_FRIENDS, parameters, error_codes=codes, splitter='|')

        ret = []
        for elem in resp:
            temp = mapper_util.map(elem.split(':'))

            parse_dict = {
                'name': temp[i.USER_NAME],
                'id': temp[i.USER_PLAYER_ID],
                'account_id': temp[i.USER_ACCOUNT_ID]
            }

            ret.append(
                ClassConverter.abstractuser_convert(parse_dict, client)
            )

        return ret


    async def get_top(self, strategy, count: int, *, client):
        from .classconverter import ClassConverter

        needs_login = (strategy.value in (1, 2))

        # special case: map 'players' -> 'top'
        strategy = strategy.name.lower() if strategy.value else 'top'

        params = Params().create_new().put_type(strategy).put_count(count)
        codes = {
            -1: MissingAccess(message='Failed to fetch leaderboard for strategy: {!r}.'.format(strategy))
        }

        if needs_login:
            check.is_logged_obj(client, 'get_top')
            params.put_definer('accountid', client.account_id).put_password(client.encodedpass)

        parameters = params.finish()

        resp = await http.request(Route.GET_USER_TOP, parameters, error_codes=codes, splitter='|')

        res = []
        for data in filter(_is_not_empty, resp):
            mapped = mapper_util.map(data.split(':'))
            stats = ClassConverter.user_stats_convert(mapped, client)
            res.append(stats)

        return res

    async def test_captcha(self):
        image_bytes = await http.request(Route.CAPTCHA)
        res = await Captcha.aio_solve(image_bytes, should_log=True)
        return res


    async def login(self, client, user: str, password: str):
        parameters = (
            Params().create_new().put_login_definer(username=user, password=password)
            .finish_login()
        )
        codes = {
            -1: LoginFailure(login=user, password=password)
        }

        resp = await http.request(Route.LOGIN, parameters, splitter=',', error_codes=codes)

        account_id, id = resp

        prepared = {
            'name': user, 'password': password,
            'account_id': int(account_id), 'id': int(id)
        }
        for attr, value in prepared.items():
            client._upd(attr, value)


    async def get_account_url(self, user_or_id, *, type: str = 'load'):
        cases = {
            'save': 1, 'load': 2
        }
        type = cases.get(type)

        # add both int and user-objects support
        id = user_or_id if not hasattr(user_or_id, 'account_id') else user_or_id.account_id

        parameters = Params().create_new('web').put_definer('accountid', id).put_type(type).finish()

        resp = await http.request(Route.GET_ACCOUNT_URL, parameters)

        if not isinstance(resp, str):  # not link?
            raise MissingAccess(message='Unknown response when searching for account url: {!r}'.format(resp))

        link = resp + '/database/'  # http://ip/database/
        return link

    async def load_save(self, client):
        link = await client.get_account_url()

        parameters = (
            Params().create_new().put_username(client.name).put_definer('password', client.password)
            .finish_login()
        )
        codes = {
            -11: MissingAccess(message='Failed to load data for user: {!r}.'.format(client))
        }

        resp = await http.request(Route.LOAD_DATA, parameters, error_codes=codes, custom_base=link)

        try:
            raw_save = Coder.decode_gamesave(resp, needs_xor=False)
            save = SaveParser.parse(raw_save)

            client._upd('raw_save', raw_save)
            client._upd('save', save)

            return True

        except Exception:
            return False


    async def search_levels_on_page(
        self, page: int = 0, query: str = '', filters: Filters = None,
        user=None, gauntlet: int = None, *, raise_errors: bool = True, client
    ):
        from .classconverter import ClassConverter

        if filters is None:
            filters = Filters.setup_empty()

        params = (
            Params().create_new().put_definer('search', query)
            .put_page(page).put_total(0).put_filters(filters)
        )
        codes = {
            -1: MissingAccess(message='No levels were found.')
        }
        if filters.strategy == SearchStrategy.BY_USER:

            if user is None:
                check.is_logged_obj(client, 'search_levels_on_page(...)')

                id = client.id

                params.put_definer('accountid', client.account_id).put_password(client.encodedpass)
                params.put_local(1)

            else:
                id = user if isinstance(user, int) else user.id

            params.put_definer('search', id)  # override the 'str' parameter in request

        elif filters.strategy == SearchStrategy.FRIENDS:
            check.is_logged_obj(client, 'search_levels_on_page(..., client=client)')
            params.put_definer('accountid', client.account_id).put_password(client.encodedpass)

        if gauntlet is not None:
            params.put_definer('gauntlet', gauntlet)

        parameters = params.finish()

        resp = await http.request(
            Route.LEVEL_SEARCH, parameters, raise_errors=raise_errors,
            error_codes=codes, splitter='#')

        if not resp:
            return list()

        lvdata, cdata, sdata = resp[:3]

        songs = []
        for s in filter(_is_not_empty, sdata.split('~:~')):
            song = ClassConverter.song_convert(mapper_util.map(s.split('~|~')))
            songs.append(song)

        creators = []
        for c in filter(_is_not_empty, cdata.split('|')):
            creator = ClassConverter.abstractuser_convert(
                {k: v for k, v in zip(('id', 'name', 'account_id'), c.split(':'))}, client=client
            )
            creators.append(creator)

        levels = []
        ext = ['101', '0', '102', '-1', '103', '-1']

        for lv in filter(_is_not_empty, lvdata.split('|')):
            data = mapper_util.map(lv.split(':') + ext)

            song_id = data.get(i.LEVEL_SONG_ID)
            song = Converter.to_normal_song(
                data.get(i.LEVEL_AUDIO_TRACK)
            ) if not song_id else utils.get(songs, id=song_id)

            creator_id = data.get(i.LEVEL_CREATOR_ID)
            creator = utils.get(creators, id=creator_id)
            if creator is None:
                creator = UnregisteredUser(creator_id)

            levels.append(ClassConverter.level_convert(data, song, creator, client))

        return levels


    async def search_levels(self, query: str = '', filters: Filters = None, user=None,
        pages: Sequence[int] = None, *, client
    ):
        to_run = [
            self.search_levels_on_page(
                query=query, filters=filters, user=user, page=page, raise_errors=False, client=client
            ) for page in pages
        ]

        return await self.run_many(to_run)


    async def report_level(self, level):
        parameters = Params().create_new('web').put_definer('levelid', level.id).finish()
        codes = {
            -1: MissingAccess(message='Failed to report a level: {!r}.'.format(level))
        }

        await http.request(Route.REPORT_LEVEL, parameters, error_codes=codes)

    async def delete_level(self, level, *, client):
        parameters = (
            Params().create_new().put_definer('accountid', client.account_id)
            .put_definer('levelid', level.id).put_password(client.encodedpass).finish_level()
        )

        resp = await http.request(Route.DELETE_LEVEL, parameters)

        if resp != 1:
            raise MissingAccess(message='Failed to delete a level: {}.'.format(level))

        # update level's is_alive coroutine to return False only.
        async def is_alive(*args):
            return False

        level.is_alive = is_alive


    async def update_level_desc(self, level, content, *, client):
        parameters = (
            Params().create_new().put_definer('accountid', client.account_id)
            .put_password(client.encodedpass).put_definer('levelid', level.id)
            .put_level_desc(content).finish()
        )

        resp = await http.request(Route.UPDATE_LEVEL_DESC, parameters)

        if resp != 1:
            raise MissingAccess(message='Failed to update description of the level: {}.'.format(level))

        # update level's description on success
        level.options['description'] = content


    async def rate_level(self, level, rating: int, *, client):
        assert 0 < rating <= 10, 'Invalid star value given.'

        rs = Coder.gen_rs()
        values = [level.id, rating, rs, client.account_id, 0, 0]
        chk = Coder.gen_chk(type='like_rate', values=values)

        parameters = (
            Params().create_new().put_definer('levelid', level.id)
            .put_definer('accountid', client.account_id).put_password(client.encodedpass)
            .put_udid(0).put_uuid(0).put_definer('stars', rating).put_rs(rs).put_chk(chk).finish()
        )

        resp = await http.request(Route.RATE_LEVEL_STARS, parameters)

        if resp != 1:
            raise MissingAccess(message='Failed to rate level: {}.'.format(level))


    async def rate_demon(self, level, demon_rating, mod: bool, *, client):
        rating_level = demon_rating.value

        parameters = (
            Params().create_new().put_definer('accountid', client.account_id)
            .put_password(client.encodedpass).put_definer('levelid', level.id)
            .put_definer('rating', rating_level).put_mode(int(mod)).finish_mod()
        )
        codes = {
            -2: MissingAccess(message='Attempt to rate as mod without mod permissions.')
        }

        resp = await http.request(Route.RATE_LEVEL_DEMON, parameters, error_codes=codes)

        if not resp:
            return False
        elif isinstance(resp, int) and resp >= 0:
            return True


    async def send_level(self, level, rating: int, featured: bool, *, client):
        parameters = (
            Params().create_new().put_definer('accountid', client.account_id)
            .put_password(client.encodedpass).put_definer('levelid', level.id)
            .put_definer('stars', rating).put_feature(int(featured)).finish_mod()
        )
        codes = {
            -2: MissingAccess(message='Missing moderator permissions to send a level: {!r}.'.format(level))
        }

        resp = await http.request(Route.SUGGEST_LEVEL_STARS, parameters, error_codes=codes)

        if resp != 1:
            raise MissingAccess(message='Failed to send a level: {!r}.'.format(level))


    async def like(self, item, dislike: bool = False, *, client):
        if hasattr(item, 'is_featured'):  # level
            typeid = 1
            special = 0

        elif hasattr(item, 'is_spam'):  # comment
            if not item.type.value:  # level comment
                typeid = 2
                special = item.level_id

            else:  # profile comment
                typeid = 3
                special = item.id

        else:  # wrong type?!
            return

        like = dislike ^ 1

        rs = Coder.gen_rs()
        values = [special, item.id, typeid, like, rs, client.account_id, 0, 0]
        chk = Coder.gen_chk(type='like_rate', values=values)

        parameters = (
            Params().create_new().put_definer('accountid', client.account_id)
            .put_password(client.encodedpass).put_udid(0).put_uuid(0)
            .put_definer('itemid', item.id).put_like(like).put_type(typeid)
            .put_special(special).put_rs(rs).put_chk(chk).finish()
        )

        resp = await http.request(Route.LIKE_ITEM, parameters)

        if resp != 1:
            raise MissingAccess(message='Failed to like an item: {}.'.format(item))


    async def get_page_messages(
        self, sent_or_inbox: str, page: int, *, raise_errors: bool = True, client
    ):
        from .classconverter import ClassConverter

        assert sent_or_inbox in ('inbox', 'sent')
        inbox = 0 if sent_or_inbox != 'sent' else 1

        parameters = (
            Params().create_new().put_definer('accountid', client.account_id)
            .put_password(client.encodedpass).put_page(page).put_total(0).get_sent(inbox).finish()
        )
        codes = {
            -1: MissingAccess(message='Failed to get messages.'),
            -2: NothingFound('gd.Message')
        }

        resp = await http.request(
            Route.GET_PRIVATE_MESSAGES, parameters, error_codes=codes,
            splitter_func=lambda s: s.split('#')[0].split('|'), raise_errors=raise_errors
        )
        if resp is None:
            return list()

        res = []
        for elem in resp:
            mapped = mapper_util.map(elem.split(':'))
            res.append(
                ClassConverter.message_convert(
                    mapped, client.get_parse_dict(), client
                )
            )

        return res


    async def get_messages(self, sent_or_inbox: str, pages: Sequence[int] = None, *, client):
        assert sent_or_inbox in ('inbox', 'sent')

        to_run = [
            self.get_page_messages(
                sent_or_inbox=sent_or_inbox, page=page, raise_errors=False, client=client
            ) for page in pages
        ]

        return await self.run_many(to_run)


    async def post_comment(self, content: str, *, client):
        to_gen = [client.name, 0, 0, 1]

        parameters = (
            Params().create_new().put_definer('accountid', client.account_id)
            .put_username(client.name).put_password(client.encodedpass)
            .put_comment(content, to_gen).comment_for('profile').finish()
        )
        codes = {
            -1: MissingAccess(message='Failed to post a comment.')
        }

        await http.request(Route.UPLOAD_ACC_COMMENT, parameters, error_codes=codes)


    async def comment_level(self, level, content: str, percentage: int, *, client):
        assert percentage <= 100, '{}% > 100% percentage arg was recieved.'.format(percentage)

        percentage = round(percentage)  # just in case
        to_gen = [client.name, level.id, percentage, 0]

        parameters = (
            Params().create_new().put_definer('accountid', client.account_id)
            .put_username(client.name).put_password(client.encodedpass)
            .put_comment(content, to_gen).comment_for('level', level.id)
            .put_percent(percentage).finish()
        )
        codes = {
            -1: MissingAccess(message='Failed to post a comment on a level: {!r}.'.format(level))
        }

        await http.request(Route.UPLOAD_COMMENT, parameters, error_codes=codes)


    async def edit(self, name: str, password: str, *, client):
        _, cookie = await http.request(Route.MANAGE_ACCOUNT, get_cookies=True)
        captcha = await http.request(Route.CAPTCHA, cookie=cookie)
        number = await Captcha.aio_solve(captcha, should_log=True)
        parameters = (
            Params().create_new('web').put_for_management(
                client.name, client.password, number).close()
        )
        await http.request(Route.MANAGE_ACCOUNT, parameters, cookie=cookie)

        if name is not None:
            parameters = Params().create_new('web').put_for_username(client.name, name).close()
            resp = await http.request(Route.CHANGE_USERNAME, parameters, cookie=cookie)
            if ('Your username has been changed to' in resp):
                client._upd('name', name)
            else:
                raise FailedToChange('name')

        if password is not None:
            await http.request(Route.CHANGE_PASSWORD, cookie=cookie)
            parameters = (
                Params().create_new('web').put_for_password(
                    client.name, client.password, password).close()
            )
            resp = await http.request(Route.CHANGE_PASSWORD, parameters, cookie=cookie)
            if ('Password change failed' in resp):
                raise FailedToChange('password')
            else:
                client._upd('password', password)


    async def delete_comment(self, comment, *, client):
        cases = {
            0: Route.DELETE_LEVEL_COMMENT,
            1: Route.DELETE_ACC_COMMENT
        }
        route = cases.get(comment.type.value)
        parameters = (
            Params().create_new().put_definer('commentid', comment.id)
            .put_definer('accountid', client.account_id).put_password(client.encodedpass)
            .comment_for(comment.type.name.lower(), comment.level_id).finish()
        )
        resp = await http.request(route, parameters)
        if resp != 1:
            raise MissingAccess(message='Failed to delete a comment: {!r}.'.format(comment))


    async def send_friend_request(self, target, message, client):
        if message is None:
            message = ''

        parameters = (
            Params().create_new().put_definer('accountid', client.account_id)
            .put_recipient(target.account_id).put_fr_comment(message)
            .put_password(client.encodedpass).finish()
        )
        resp = await http.request(Route.SEND_REQUEST, parameters)

        if not resp:  # if request is already sent
            return

        elif resp != 1:
            raise MissingAccess(message='Failed to send a friend request to user: {!r}.'.format(target))


    async def delete_friend_req(self, req):
        client = req._client
        user = req.author if not req.type.value else req.recipient
        parameters = (
            Params().create_new().put_definer('accountid', client.account_id)
            .put_definer('user', user.account_id).put_password(client.encodedpass)
            .put_is_sender(req.type.name.lower()).finish()
        )
        resp = await http.request(Route.DELETE_REQUEST, parameters)
        if resp != 1:
            raise MissingAccess(message='Failed to delete a friend request: {!r}.'.format(req))


    async def accept_friend_req(self, req):
        client = req._client
        if req.type.value:  # is gd.MessageOrRequestType.SENT
            raise MissingAccess(
                message="Failed to accept a friend request. Reason: request is sent, not recieved one."
            )
        parameters = (
            Params().create_new().put_definer('accountid', client.account_id)
            .put_password(client.encodedpass).put_definer('user', req.author.account_id)
            .put_definer('requestid', req.id).finish()
        )
        resp = await http.request(Route.ACCEPT_REQUEST, parameters)
        if resp != 1:
            raise MissingAccess(message='Failed to accept a friend request: {!r}.'.format(req))


    async def read_friend_req(self, req):
        client = req._client
        parameters = (
            Params().create_new().put_definer('accountid', client.account_id)
            .put_password(client.encodedpass).put_definer('requestid', req.id).finish()
        )
        resp = await http.request(Route.READ_REQUEST, parameters)
        if resp != 1:
            raise MissingAccess(message='Failed to read a friend request: {!r}.'.format(req))
        req.options.update({'is_read': True})


    async def read_message(self, msg):
        client = msg._client
        parameters = (
            Params().create_new().put_definer('accountid', client.account_id)
            .put_definer('messageid', msg.id).put_is_sender(msg.type.name.lower())
            .put_password(client.encodedpass).finish()
        )
        codes = {
            -1: MissingAccess(message='Failed to read a message: {!r}.'.format(msg))
        }
        resp = await http.request(
            Route.READ_PRIVATE_MESSAGE, parameters, error_codes=codes,
            splitter=':', should_map=True
        )

        ret = Coder.decode(
            type='message', string=mapper_util.normalize(resp.get(i.MESSAGE_BODY))
        )
        msg._body = ret
        return ret


    async def delete_message(self, msg):
        client = msg._client
        parameters = (
            Params().create_new().put_definer('accountid', client.account_id)
            .put_definer('messageid', msg.id).put_password(client.encodedpass)
            .put_is_sender(msg.type.name.lower()).finish()
        )
        resp = await http.request(Route.DELETE_PRIVATE_MESSAGE, parameters)
        if resp != 1:
            raise MissingAccess(message='Failed to delete a message: {!r}.'.format(msg))


    async def get_gauntlets(self, *, client):
        from .classconverter import ClassConverter

        parameters = Params().create_new().finish()

        resp = await http.request(
            Route.GET_GAUNTLETS, parameters, splitter_func=lambda s: s.split('#')[0].split('|')
        )

        res = []
        for gdata in filter(_is_not_empty, resp):
            mapped = mapper_util.map(gdata.split(':'))
            res.append(
                ClassConverter.gauntlet_convert(mapped, client)
            )

        return res


    async def get_page_map_packs(self, page: int = 0, *, raise_errors: bool = True, client):
        from .classconverter import ClassConverter

        parameters = Params().create_new().put_page(page).finish()

        resp = await http.request(
            Route.GET_MAP_PACKS, parameters, splitter_func=lambda s: s.split('#')[0].split('|')
        )

        if resp and not resp[0]:
            if raise_errors:
                raise NothingFound('gd.MapPack')
            return list()

        res = []
        for elem in resp:
            mapped = mapper_util.map(elem.split(':'))
            res.append(
                ClassConverter.map_pack_convert(mapped, client)
            )

        return res


    async def get_map_packs(self, pages: Sequence[int] = None, *, client):
        to_run = [
            self.get_page_map_packs(
                page=page, raise_errors=False, client=client
            ) for page in pages
        ]

        return await self.run_many(to_run)


    async def get_page_friend_requests(self, sent_or_inbox: str = 'inbox',
        page: int = 0, *, raise_errors: bool = True, client):
        from .classconverter import ClassConverter

        inbox = 1 if sent_or_inbox == 'sent' else 0

        parameters = (
            Params().create_new().put_definer('accountid', str(client.account_id))
            .put_password(client.encodedpass).put_page(page).put_total(0).get_sent(inbox).finish()
        )
        codes = {
            -1: MissingAccess(message='Failed to get friend requests on page {}.'.format(page)),
            -2: NothingFound('gd.FriendRequest')
        }

        resp = await http.request(
            Route.GET_FRIEND_REQUESTS, parameters, error_codes=codes, raise_errors=raise_errors,
            splitter_func=lambda s: s.split('#')[0].split('|')
        )

        if not resp:
            return list()

        res = []
        for elem in resp:
            mapped = mapper_util.map(
                elem.split(':') + ['101', str(inbox)]
            )
            res.append(
                ClassConverter.request_convert(
                    mapped, client.get_parse_dict(), client
                )
            )

        return res


    async def get_friend_requests(
        self, sent_or_inbox: str = 'inbox', pages: Sequence[int] = None, *, client
    ):
        assert sent_or_inbox in ('sent', 'inbox')

        to_run = [
            self.get_page_friend_requests(
                sent_or_inbox=sent_or_inbox, page=page, raise_errors=False, client=client
            ) for page in pages
        ]

        return await self.run_many(to_run)


    async def retrieve_page_comments(
        self, user, type: str = 'profile', page: int = 0, *, raise_errors: bool = True, strategy
    ):
        from .classconverter import ClassConverter

        assert isinstance(page, int) and page >= 0
        assert type in ("profile", "level")

        is_level = (type == "level")

        typeid = 0 if is_level else 1
        definer = "userid" if is_level else "accountid"
        selfid = user.id if is_level else user.account_id
        route = Route.GET_COMMENT_HISTORY if is_level else Route.GET_ACC_COMMENTS

        def func(elem):
            if is_level:
                return elem.split(':')[0].split('~')
            return elem.split('~')

        param_obj = Params().create_new().put_definer(definer, selfid).put_page(page).put_total(0)
        if is_level:
            param_obj.put_mode(strategy.value)
        parameters = param_obj.finish()

        resp = await http.request(route, parameters, splitter='#')
        thing = resp[0]

        if not thing:
            if raise_errors:
                raise NothingFound('gd.Comment')
            return list()

        to_map = thing.split('|')

        res = []
        for elem in to_map:
            prepared = mapper_util.map(
                func(elem) + ['101', str(typeid)]
            )
            res.append(
                ClassConverter.comment_convert(
                    prepared, user._dict_for_parse, user._client
                )
            )

        return res


    async def retrieve_comments(
        self, user, type: str = 'profile', pages: Sequence[int] = None, *, strategy
    ):
        assert type in ('profile', 'level')

        to_run = [
            self.retrieve_page_comments(
                type=type, user=user, page=page, raise_errors=False, strategy=strategy
            ) for page in pages
        ]

        return await self.run_many(to_run)


    async def get_level_comments(self, level, strategy, amount: int):
        from .classconverter import ClassConverter

        st_value = strategy.value

        parameters = (
            Params().create_new().put_definer('levelid', level.id).put_page(0)
            .put_total(0).put_mode(st_value).put_count(amount).finish()
        )
        codes = {
            -1: MissingAccess(message='Failed to get comments of a level: {!r}.'.format(level)),
            -2: NothingFound('gd.Comment')
        }

        resp = await http.request(
            Route.GET_COMMENTS, parameters, error_codes=codes,
            splitter_func=lambda s: s.split('#')[0].split('|')
        )

        res = []
        for elem in resp:
            com_data, user_data = (mapper_util.map(part.split('~')) for part in elem.split(':'))
            com_data.update({1: level.id, 101: 0, 102: 0})

            user_dict = {
                'account_id': user_data[i.USER_ACCOUNT_ID],
                'id': com_data[i.COMMENT_AUTHOR_ID],
                'name': user_data[i.USER_NAME]
            }

            res.append(
                ClassConverter.comment_convert(
                    com_data, user_dict, level._client
                )
            )

        return res


    async def block_user(self, user, unblock: bool = False, *, client):
        route = Route.UNBLOCK_USER if unblock else Route.BLOCK_USER
        parameters = (
            Params().create_new().put_definer('accountid', client.account_id)
            .put_password(client.encodedpass)
            .put_definer('user', user.account_id).finish()
        )
        resp = await http.request(route, parameters)
        if resp != 1:
            raise MissingAccess(message='Failed to (un)block a user: {!r}.'.format(user))


    async def unfriend_user(self, user, *, client):
        parameters = (
            Params().create_new().put_definer('accountid', client.account_id)
            .put_password(client.encodedpass).put_definer('user', user.account_id).finish()
        )
        resp = await http.request(Route.REMOVE_FRIEND, parameters)
        if resp != 1:
            raise MissingAccess(message='Failed to unfriend a user: {!r}.'.format(user))


    async def send_message(self, target, subject: str, body: str, *, client):
        parameters = (
            Params().create_new().put_definer('accountid', client.account_id)
            .put_message(subject, body).put_recipient(target.account_id)
            .put_password(client.encodedpass).finish()
        )
        resp = await http.request(Route.SEND_PRIVATE_MESSAGE, parameters)
        if resp != 1:
            raise MissingAccess(message='Failed to send a message to a user: {!r}.'.format(target))


    async def update_settings(
        self, msg: int, friend_req: int, comments: int,
        youtube: str, twitter: str, twitch: str, *, client
    ):
        parameters = (
            Params().create_new('web').put_definer('accountid', client.account_id)
            .put_password(client.encodedpass)
            .put_profile_upd(msg, friend_req, comments, youtube, twitter, twitch).finish_login()
        )
        resp = await http.request(Route.UPDATE_ACC_SETTINGS, parameters)
        if resp != 1:
            raise MissingAccess(message='Failed to update profile settings of a client: {!r}.'.format(client))


    async def run_many(self, tasks):
        res = await asyncio.gather(*tasks)

        res = [elem for elem in res if elem]

        if all(isinstance(elem, list) for elem in res):
            res = list(itertools.chain.from_iterable(res))

        return res


def _is_not_empty(sequence):
    return bool(len(sequence))

_session = GDSession()



class Route:
    BLOCK_USER = 'blockGJUser20'
    UNBLOCK_USER = 'unblockGJUser20'
    RATE_LEVEL_STARS = 'rateGJStars211'
    RATE_LEVEL_DEMON = 'rateGJDemon21'
    SUGGEST_LEVEL_STARS = 'suggestGJStars20'
    GET_USER_INFO = 'getGJUserInfo20'
    USER_SEARCH = 'getGJUsers20'
    GET_USER_TOP = 'getGJScores20'
    UPDATE_USER_SCORE = 'updateGJUserScore22'
    REPORT_LEVEL = 'reportGJLevel'
    DOWNLOAD_LEVEL = 'downloadGJLevel22'
    LEVEL_SEARCH = 'getGJLevels21'
    UPDATE_LEVEL_DESC = 'updateGJDesc20'
    GET_MAP_PACKS = 'getGJMapPacks21'
    GET_GAUNTLETS = 'getGJGauntlets21'
    DELETE_LEVEL = 'deleteGJLevelUser20'
    GET_SONG_INFO = 'getGJSongInfo'
    GET_FRIENDS = 'getGJUserList20'
    REMOVE_FRIEND = 'removeGJFriend20'
    GET_ACC_COMMENTS = 'getGJAccountComments20'
    GET_COMMENTS = 'getGJComments21'
    GET_PRIVATE_MESSAGES = 'getGJMessages20'
    READ_PRIVATE_MESSAGE = 'downloadGJMessage20'
    SEND_PRIVATE_MESSAGE = 'uploadGJMessage20'
    DELETE_PRIVATE_MESSAGE = 'deleteGJMessages20'
    GET_TIMELY = 'getGJDailyLevel'
    LOGIN = 'accounts/loginGJAccount'
    LOAD_DATA = 'accounts/syncGJAccountNew'
    CAPTCHA = 'accounts/captcha'
    MANAGE_ACCOUNT = 'accounts/accountManagement'
    CHANGE_PASSWORD = 'accounts/changepassword'
    CHANGE_USERNAME = 'accounts/changeusername'
    DELETE_ACC_COMMENT = 'deleteGJAccComment20'
    DELETE_LEVEL_COMMENT = 'deleteGJComment20'
    UPLOAD_COMMENT = 'uploadGJComment21'
    UPLOAD_ACC_COMMENT = 'uploadGJAccComment20'
    UPDATE_ACC_SETTINGS = 'updateGJAccSettings20'
    GET_FRIEND_REQUESTS = 'getGJFriendRequests20'
    READ_REQUEST = 'readGJFriendRequest20'
    DELETE_REQUEST = 'deleteGJFriendRequests20'
    ACCEPT_REQUEST = 'acceptGJFriendRequest20'
    SEND_REQUEST = 'uploadFriendRequest20'
    REQUEST_MODERATOR = 'requestUserAccess'
    GET_COMMENT_HISTORY = 'getGJCommentHistory'
    GET_ACCOUNT_URL = 'getAccountURL'
    LIKE_ITEM = 'likeGJItem211'
    NEWGROUNDS_SONG_LISTEN = 'https://www.newgrounds.com/audio/listen/'



from collections import namedtuple
import re
# more things soon <3

Save = namedtuple('Save', 'completed followed')

class SaveParser:
    @classmethod
    def parse(cls, xml):
        completed = cls.parse_completed_levels(xml)
        followed = cls.parse_followed(xml)
        return Save(completed=completed, followed=followed)

    @classmethod
    def parse_completed_levels(cls, xml):
        pattern = '<k>c_(\d+)</k><s>1</s>'
        found = re.findall(pattern, xml)

        return map_to_int(found)

    @classmethod
    def parse_followed(cls, xml):
        try:
            section = slice_between(xml, 'GLM', 6, 8)
        except Exception:
            section = slice_between(xml, 'GLM', 6, 7)

        pattern = '<k>(\d+)</k><s>1</s>'
        found = re.findall(pattern, section)

        return map_to_int(found)

def map_to_int(seq):
    return [int(x) for x in seq]

def slice_between(sequence, section_name, x, y):
    fix = len(section_name) + 3

    x, y = ('{}_{:02}'.format(section_name, z) for z in (x, y))
    i, j = (sequence.index(z) for z in (x, y))
    i += fix

    return sequence[i:j]



"""Module that implements some Geometry Dash jokes..."""

from .utils.wrap_tools import make_repr


class JokeNotFoundError(AttributeError):
    def __init__(self, name):
        message = 'No joke was found under name: {!r}.'.format(name)
        super().__init__(message)


class JokeHandler:
    def __init__(self):
        self.jokes = {
            'anime': 'I see, you are a person of culture.',
            'blaze': 'What if we would send all levels?',
            'colon': 'Did you use the right pusab?',
            'etzer': 'Have you sent it to RobTop yet? 8)',
            'michigun': 'Every level needs a triple... ΔΔΔ',
            'nekit': '<sarcasm> HTML is a programming language. </sarcasm>',
            'owo': 'Hewwo, my fwiend! owo',
            'serponge': 'New AlterGame in progress :pog:',
            'viprin': '[Ctrl+D] | [Ctrl+C & Ctrl+V]'
        }

    def __repr__(self):
        info = {
            key: repr(value) for key, value in self.jokes.items()
        }
        return make_repr(self, info)

    def __getattr__(self, attr):
        name = attr.lower()

        if name not in self.jokes:
            raise JokeNotFoundError(attr)

        joke = str(self.jokes.get(name))
        print(joke)


jokes = JokeHandler()



import asyncio
import functools

from .wrap_tools import find_subclass, add_method, del_method

__all__ = (
    'run_blocking_io', 'wait', 'run',
    'cancel_all_tasks', 'shutdown_loop',
    'coroutine'
)

coroutine = find_subclass('coroutine')


async def run_blocking_io(func, *args, **kwargs):
    """|coro|

    Run some blocking function in an event loop.

    If there is a running loop, ``'func'`` is executed in it.

    Otherwise, a new loop is being created and closed at the end of the execution.

    Example:

    .. code-block:: python3

        def make_image():
            ...  # long code of creating an image

        # somewhere in an async function:

        await run_blocking_io(make_image)
    """
    loop = asyncio._get_running_loop()
    should_close = False

    if loop is None:
        loop = asyncio.new_event_loop()
        should_close = True

    try:
        return await loop.run_in_executor(None, functools.partial(func, *args, **kwargs))

    finally:
        if should_close:
            shutdown_loop(loop)


async def wait(fs, *, loop=None, timeout=None, return_when='ALL_COMPLETED'):
    """A function that is calling :func:`asyncio.wait`.

    Used for less imports inside and outside of this library.

    Wait for the Futures and coroutines given by fs to complete.

    The sequence futures must not be empty.

    Coroutines will be wrapped in Tasks.

    Returns two sets of Future: (done, pending).

    Usage:

    .. code-block:: python3

        done, pending = await gd.utils.wait(fs)

    .. note::

        This does not raise :exc:`TimeoutError`! Futures that aren't done
        when the timeout occurs are returned in the second set.
    """
    try:
        fs = list(fs)
    except TypeError:  # not iterable
        pass
    return await asyncio.wait(fs, loop=loop, timeout=timeout, return_when=return_when)


def run(coro, *, loop=None, debug: bool = False):
    """Run a |coroutine_link|_.

    This function runs the passed coroutine, taking care
    of the event loop and shutting down asynchronous generators.

    This function is basically ported from Python 3.7 for backwards compability
    with earlier versions of Python.

    This function cannot be called when another event loop is
    running in the same thread.

    If ``debug`` is ``True``, the event loop will be run in debug mode.

    This function creates a new event loop and closes it at the end if a ``loop`` is ``None``.

    If a loop is given, this function basically calls :meth:`asyncio.AbstractEventLoop.run_until_complete`.

    It should be used as a main entry point to asyncio programs, and should
    ideally be called only once.

    Example:

    .. code-block:: python3

        async def test(pid):
            return pid

        one = gd.utils.run(test(1))

    Parameters
    ----------
    coro: |coroutine_link|_
        Coroutine to run.

    loop: Optional[:class:`asyncio.AbstractEventLoop`]
        A loop to run ``coro`` with. If ``None`` or omitted, a new event loop is created.

    debug: :class:`bool`
        Whether or not to run event loop in debug mode.

    Returns
    -------
    `Any`
        Anything that ``coro`` returns.
    """

    if asyncio._get_running_loop() is not None:
        raise RuntimeError('Can not perform gd.utils.run() in a running event loop.')

    if not asyncio.iscoroutine(coro):
        raise ValueError('A coroutine was expected, got {!r}.'.format(coro))

    # if a loop is given, let's just run our coro in it
    if loop is not None:
        return loop.run_until_complete(coro)

    loop = asyncio.new_event_loop()

    try:
        asyncio.set_event_loop(loop)
        loop.set_debug(debug)
        return loop.run_until_complete(coro)

    finally:
        shutdown_loop(loop, True)


def cancel_all_tasks(loop, f_name: str = 'gd.utils.run'):
    """Cancels all tasks in a loop.

    Parameters
    ----------
    loop: :class:`asyncio.AbstractEventLoop`
        Event loop to cancel tasks in.
    """
    try:
        to_cancel = asyncio.all_tasks(loop)
    except AttributeError:  # py < 3.7
        to_cancel = asyncio.Task.all_tasks(loop)

    if not to_cancel:
        return

    for task in to_cancel:
        task.cancel()

    loop.run_until_complete(
        asyncio.gather(*to_cancel, loop=loop, return_exceptions=True)
    )

    for task in to_cancel:
        if task.cancelled():
            continue

        if task.exception() is not None:
            loop.call_exception_handler({
                'message': 'Unhandled exception during {}() shutdown'.format(f_name),
                'exception': task.exception(),
                'task': task
            })


def shutdown_loop(loop, set_event_loop_to_none: bool = False):
    try:
        cancel_all_tasks(loop)
        loop.run_until_complete(loop.shutdown_asyncgens())

    finally:
        if set_event_loop_to_none:
            asyncio.set_event_loop(None)
        loop.close()


def _run(self):
    """Run the coroutine in a new event loop,
    closing the loop after execution.
    """
    loop = asyncio.new_event_loop()

    try:
        return loop.run_until_complete(self)

    finally:
        loop.close()


def enable_run_method(on: bool = True):
    """Add or delete 'run' method of a coroutine."""
    if on:
        add_method(coroutine, _run, name='run')
    else:
        del_method(coroutine, 'run')

enable_run_method()



import ctypes
import functools
import gc
import logging
import time  # not for time.sleep() :)

from ..errors import NotLoggedError, MissingAccess

__all__ = (
    'check', 'benchmark', 'new_method', 'add_method',
    'del_method', 'make_repr', 'find_subclass', 'get_instances_of'
)

log = logging.getLogger(__name__)


def get_class_dict(cls):
    """Gets 'cls.__dict__' that can be edited."""
    return ctypes.py_object.from_address(id(cls.__dict__) + 16).value


class check:
    @classmethod
    def is_logged(cls):
        # decorator that checks if passed client is logged in.
        def decorator(func):
            @functools.wraps(func)
            def wrapper(obj, *args, **kwargs):

                client = obj if hasattr(obj, 'login') else obj._client

                if client is None:
                    raise MissingAccess(
                        message=(
                            'Attempt to check if client is logged for {!r} returned None. '
                            'Have you made this object by hand?').format(obj)
                        )

                if not client.is_logged():
                    raise NotLoggedError(func.__name__)

                return func(obj, *args, **kwargs)

            return wrapper
        return decorator

    @classmethod
    def is_logged_obj(cls, obj, func_name):
        try:
            client = obj if hasattr(obj, 'login') else obj._client

        except AttributeError:
            raise MissingAccess(
                message='None was recieved as an <obj> arg when checking in {!r}.'.format(func_name)
            ) from None

        if not client.is_logged():
            raise NotLoggedError(func_name)


def benchmark(func):
    @functools.wraps(func)
    def decorator(*args, **kwargs):

        start = time.perf_counter()
        res = func(*args, **kwargs)
        end = time.perf_counter()

        time_taken = (end-start)*1000

        thing = (
            'Executed "{}(*args, **kwargs)"\n'
            'Estimated time: {:,.2f}ms.'
        ).format(func.__name__, time_taken)

        log.debug(thing)

        return res
    return decorator


def new_method(cls):
    # decorator that adds new methods to a 'cls'.
    def decorator(func):
        cls_d = get_class_dict(cls)
        cls_d[func.__name__] = func

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper
    return decorator


def make_repr(obj, info = {}):
    """Creates a nice __repr__ for the object."""
    module = obj.__module__.split('.')[0]
    name = obj.__class__.__name__

    if not info:
        return '<{}.{}>'.format(module, name)

    info = (' '.join('%s=%s' % t for t in info.items()))

    return '<{}.{} {}>'.format(module, name, info)


def find_subclass(string_name, superclass=object):
    """Fetches subclass of 'superclass' by 'string_name'."""
    subclasses = {cls.__name__: cls for cls in superclass.__subclasses__()}
    return subclasses.get(string_name, superclass)


def get_instances_of(obj_class=object):
    objects = gc.get_objects()
    return [obj for obj in objects if isinstance(obj, obj_class)]


def del_method(cls, method_name):
    """Delete a method of a 'cls'."""
    cls_d = get_class_dict(cls)
    cls_d.pop(method_name, None)


def add_method(cls, func, *, name: str = None):
    """Adds a new method to a 'cls'."""
    cls_d = get_class_dict(cls)

    if name is None:
        name = func.__name__

    cls_d[name] = func



import functools
from enum import Enum

from typing import Union

from ..errors import FailedConversion

__all__ = (
    'NEnum',
    'IconType',
    'MessagePolicyType',
    'CommentPolicyType',
    'FriendRequestPolicyType',
    'StatusLevel',
    'LevelLength',
    'LevelDifficulty',
    'DemonDifficulty',
    'TimelyType',
    'CommentType',
    'MessageOrRequestType',
    'CommentStrategy',
    'LeaderboardStrategy',
    'SearchStrategy',
    'GauntletEnum'
)

def _name_to_enum(x: str):
    return x.upper().replace(' ', '_')


def _enum_to_name(x: Enum):
    name = x.name
    return name if name == 'XL' else name.replace('_', ' ').title()


def value_to_enum(enum: Enum, x: Union[int, str, Enum]):
    """Tries to convert given value to Enum object.

    Example:

    .. code-block:: python3

        from gd.utils import value_to_enum

        enum = gd.IconType

        value_to_enum(enum, 0)                -> gd.IconType.CUBE
        value_to_enum(enum, 'cube')           -> gd.IconType.CUBE
        value_to_enum(enum, gd.IconType.CUBE) -> gd.IconType.CUBE

    Parameters
    ----------
    enum: :class:`enum.Enum`
        Instance of *Enum* to convert ``x`` to.

    x: Union[:class:`int`, :class:`str`, :class:`enum.Enum`]
        Object to convert to ``enum``.

    Returns
    -------
    :class:`enum.Enum`
        Result of converting ``x`` to ``enum``.

    Raises
    ------
    :exc:`.FailedConversion`
        Failed to convert ``x`` to ``enum``.
    """
    try:
        # if int -> enum of value x
        if isinstance(x, int):
            return enum(x)

        # if str -> enum of name x (converted)
        elif isinstance(x, str):
            return enum[_name_to_enum(x)]

        # if enum -> enum of value x.value
        elif isinstance(x, NEnum):
            return enum(x.value)

        # let's raise it here, so if it is raised, we can tell that invalid type was given.
        raise ValueError

    except ValueError:
        raise FailedConversion(enum=enum, value=x) from None


@functools.total_ordering
class NEnum(Enum):
    """Subclass of :class:`enum.Enum`, used for creating enums in gd.py.

    .. container:: operations

        .. describe:: x == y

            Checks if two enums are equal.

        .. describe:: x != y

            Checks if two enums are not equal.

        .. describe:: x > y

            Checks if value of x is higher than value of y.
            Raises an error if values are not instances of :class:`int`.

        .. describe:: x < y

            Checks if value of x is lower than value of y.
            Raises an error on not-integer type as well.

        .. describe:: x >= y

            Same as *x == y or x > y*.

        .. describe:: x <= y

            Same as *x == y or x < y*.

        .. describe:: hash(x)

            Return the enum's hash.

        .. describe:: str(x)

            Returns :attr:`.NEnum.desc`.
    """

    def __str__(self):
        return self.desc

    def __repr__(self):
        return '<gd.{0}.{1}: {2} ({3})>'.format(
            self.__class__.__name__, self.name, self.value, self.desc
        )

    def __eq__(self, other):
        return isinstance(other, type(self)) and self.value == other.value

    def __ne__(self, other):
        return isinstance(other, type(self)) and self.value != other.value

    def __gt__(self, other):
        return isinstance(other, type(self)) and self.value > other.value

    def __hash__(self):
        return hash(self.__repr__())

    @property
    def desc(self):
        """:class:`str`: More readable version of the name, e.g.

        .. code-block:: python3

            gd.SearchStrategy.BY_USER.desc -> 'By User'.
        """
        return _enum_to_name(self)

    @classmethod
    def from_value(cls, value: Union[int, str]):
        """Returns *Enum* with given value.

        .. note::

            This can only be called in a class **derived** from :class:`.NEnum`,
            so *NEnum.from_value(...)* will raise an error.

            Example:

            .. code-block:: python3

                mod = gd.StatusLevel.from_value('moderator')

                hall_of_fame = gd.SearchStrategy.from_value('hall of fame')

        Parameters
        ----------
        value: Union[:class:`int`, :class:`str`]
            A value to make *Enum* from.

        Returns
        -------
        :class:`.NEnum`
            Enum from the :class:`.NEnum` subclass.

        Raises
        ------
        :exc:`.FailedConversion`
            Failed to convert ``value`` to *Enum*.

        :exc:`TypeError`
            Invalid *Enum* class was given.
        """
        assert cls is not NEnum
        return value_to_enum(cls, value)

class IconType(NEnum):
    """An enumeration of icon types."""
    CUBE = 0
    SHIP = 1
    BALL = 2
    UFO = 3
    WAVE = 4
    ROBOT = 5
    SPIDER = 6


class MessagePolicyType(NEnum):
    """An enumeration for message policy."""
    OPENED_TO_ALL = 0
    OPENED_TO_FRIENDS_ONLY = 1
    CLOSED = 2


class CommentPolicyType(NEnum):
    """An enumeration for comment policy."""
    OPENED_TO_ALL = 0
    OPENED_TO_FRIENDS_ONLY = 1
    CLOSED = 2


class FriendRequestPolicyType(NEnum):
    """An enumeration for friend request policy."""
    OPENED = 0
    CLOSED = 1


class StatusLevel(NEnum):
    """An enumeration for Geometry Dash Status."""
    USER = 0
    MODERATOR = 1
    ELDER_MODERATOR = 2


class LevelLength(NEnum):
    """An enumeration for level lengths."""
    TINY = 0
    SHORT = 1
    MEDIUM = 2
    LONG = 3
    XL = 4


class LevelDifficulty(NEnum):
    """An enumeration for level difficulties."""
    NA = -1
    AUTO = -3
    EASY = 1
    NORMAL = 2
    HARD = 3
    HARDER = 4
    INSANE = 5
    DEMON = -2


class DemonDifficulty(NEnum):
    """An enumeration for demon difficulties."""
    EASY_DEMON = 1
    MEDIUM_DEMON = 2
    HARD_DEMON = 3
    INSANE_DEMON = 4
    EXTREME_DEMON = 5


class TimelyType(NEnum):
    """An enumeration for timely types."""
    NOT_TIMELY = 0
    DAILY = 1
    WEEKLY = 2


class CommentType(NEnum):
    """An enumeration for comment objects."""
    LEVEL = 0
    PROFILE = 1


class MessageOrRequestType(NEnum):
    """An enumeration for message and friend request objects."""
    NORMAL = 0
    SENT = 1


class CommentStrategy(NEnum):
    """An enumeration for comment searching."""
    RECENT = 0
    MOST_LIKED = 1


class LeaderboardStrategy(NEnum):
    """An enumeration for getting leaderboard users."""
    PLAYERS = 0
    FRIENDS = 1
    RELATIVE = 2
    CREATORS = 3


class GauntletEnum(NEnum):
    """An enumeration for gauntlets."""
    FIRE = 1
    ICE = 2
    POISON = 3
    SHADOW = 4
    LAVA = 5
    BONUS = 6
    CHAOS = 7
    DEMON = 8
    TIME = 9
    CRYSTAL = 10
    MAGIC = 11
    SPIKE = 12
    MONSTER = 13
    DOOM = 14
    DEATH = 15


class SearchStrategy(NEnum):
    """An enumeration for search strategy."""
    REGULAR = 0
    MOST_DOWNLOADED = 1
    MOST_LIKED = 2
    TRENDING = 3
    RECENT = 4
    BY_USER = 5
    FEATURED = 6
    MAGIC = 7
    MAP_PACKS = 10
    AWARDED = 11
    FOLLOWED = 12
    FRIENDS = 13
    HALL_OF_FAME = 16
